from __future__ import annotations

import asyncio
from asyncio import AbstractEventLoop
from typing import NoReturn

import pytest

from asphalt.core.component import (
    CLIApplicationComponent,
    Component,
    ContainerComponent,
    component_types,
)
from asphalt.core.context import Context


class DummyComponent(Component):
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.started = False

    async def start(self, ctx):
        await asyncio.sleep(0.1)
        self.started = True


@pytest.fixture(autouse=True)
def monkeypatch_plugins(monkeypatch):
    monkeypatch.setattr(component_types, "_entrypoints", {"dummy": DummyComponent})


class TestContainerComponent:
    @pytest.fixture
    def container(self) -> ContainerComponent:
        return ContainerComponent({"dummy": {"a": 1, "c": 3}})

    def test_add_component(self, container: ContainerComponent) -> None:
        """
        Test that add_component works with an without an entry point and that external
        configuration overriddes directly supplied configuration values.

        """
        container.add_component("dummy", DummyComponent, a=5, b=2)

        assert len(container.child_components) == 1
        component = container.child_components["dummy"]
        assert isinstance(component, DummyComponent)
        assert component.kwargs == {"a": 1, "b": 2, "c": 3}

    def test_add_component_with_type(self) -> None:
        """
        Test that add_component works with a `type` specified in a
        configuration overriddes directly supplied configuration values.

        """
        container = ContainerComponent({"dummy": {"type": DummyComponent}})
        container.add_component("dummy")
        assert len(container.child_components) == 1
        component = container.child_components["dummy"]
        assert isinstance(component, DummyComponent)

    @pytest.mark.parametrize(
        "alias, cls, exc_cls, message",
        [
            ("", None, TypeError, "component_alias must be a nonempty string"),
            (
                "foo",
                None,
                LookupError,
                "no such entry point in asphalt.components: foo",
            ),
            (
                "foo",
                int,
                TypeError,
                "int is not a subclass of asphalt.core.component.Component",
            ),
        ],
        ids=["empty_alias", "bogus_entry_point", "wrong_subclass"],
    )
    def test_add_component_errors(
        self,
        container: ContainerComponent,
        alias: str,
        cls: type | None,
        exc_cls: type[Exception],
        message: str,
    ) -> None:
        exc = pytest.raises(exc_cls, container.add_component, alias, cls)
        assert str(exc.value) == message

    def test_add_duplicate_component(self, container) -> None:
        container.add_component("dummy")
        exc = pytest.raises(ValueError, container.add_component, "dummy")
        assert str(exc.value) == 'there is already a child component named "dummy"'

    @pytest.mark.asyncio
    async def test_start(self, container) -> None:
        await container.start(Context())
        assert container.child_components["dummy"].started


class TestCLIApplicationComponent:
    def test_run_return_none(self, event_loop: AbstractEventLoop) -> None:
        class DummyCLIComponent(CLIApplicationComponent):
            async def run(self, ctx: Context) -> None:
                pass

        component = DummyCLIComponent()
        event_loop.run_until_complete(component.start(Context()))
        exc = pytest.raises(SystemExit, event_loop.run_forever)
        assert exc.value.code == 0

    def test_run_return_5(self, event_loop: AbstractEventLoop) -> None:
        class DummyCLIComponent(CLIApplicationComponent):
            async def run(self, ctx: Context) -> int:
                return 5

        component = DummyCLIComponent()
        event_loop.run_until_complete(component.start(Context()))
        exc = pytest.raises(SystemExit, event_loop.run_forever)
        assert exc.value.code == 5

    def test_run_return_invalid_value(self, event_loop: AbstractEventLoop) -> None:
        class DummyCLIComponent(CLIApplicationComponent):
            async def run(self, ctx: Context) -> int:
                return 128

        component = DummyCLIComponent()
        event_loop.run_until_complete(component.start(Context()))
        with pytest.warns(UserWarning) as record:
            exc = pytest.raises(SystemExit, event_loop.run_forever)

        assert exc.value.code == 1
        assert len(record) == 1
        assert str(record[0].message) == "exit code out of range: 128"

    def test_run_return_invalid_type(self, event_loop: AbstractEventLoop) -> None:
        class DummyCLIComponent(CLIApplicationComponent):
            async def run(self, ctx: Context) -> int:
                return "foo"  # type: ignore[return-value]

        component = DummyCLIComponent()
        event_loop.run_until_complete(component.start(Context()))
        with pytest.warns(UserWarning) as record:
            exc = pytest.raises(SystemExit, event_loop.run_forever)

        assert exc.value.code == 1
        assert len(record) == 1
        assert str(record[0].message) == "run() must return an integer or None, not str"

    def test_run_exception(self, event_loop: AbstractEventLoop) -> None:
        class DummyCLIComponent(CLIApplicationComponent):
            async def run(self, ctx: Context) -> NoReturn:
                raise Exception("blah")

        component = DummyCLIComponent()
        event_loop.run_until_complete(component.start(Context()))
        exc = pytest.raises(SystemExit, event_loop.run_forever)
        assert exc.value.code == 1



"""This is the server code for the Asphalt echo server tutorial."""
from asyncio import StreamReader, StreamWriter, start_server

from asphalt.core import Component, Context, run_application


async def client_connected(reader: StreamReader, writer: StreamWriter) -> None:
    message = await reader.readline()
    writer.write(message)
    writer.close()
    print("Message from client:", message.decode().rstrip())


class ServerComponent(Component):
    async def start(self, ctx: Context) -> None:
        await start_server(client_connected, "localhost", 64100)


if __name__ == "__main__":
    component = ServerComponent()
    run_application(component)



#!/usr/bin/env python3
import importlib.metadata

from packaging.version import parse

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx_autodoc_typehints",
    "sphinxcontrib.asyncio",
]

templates_path = ["_templates"]
source_suffix = ".rst"
master_doc = "index"
project = "The Asphalt Framework (core)"
author = "Alex Grönholm"
copyright = "2015, " + author

v = parse(importlib.metadata.version("asphalt"))
version = v.base_version
release = v.public

language = None

exclude_patterns = ["_build"]
pygments_style = "sphinx"
highlight_language = "python3"
todo_include_todos = False

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
htmlhelp_basename = "asphaltdoc"

intersphinx_mapping = {"python": ("http://docs.python.org/3/", None)}



import re
import sys

version_re = re.compile(r"_py(\d)(\d)\.py$")


def pytest_ignore_collect(path, config):
    match = version_re.search(path.basename)
    if match:
        version = tuple(int(x) for x in match.groups())
        if sys.version_info < version:
            return True



from __future__ import annotations

__all__ = ("executor",)

import inspect
import sys
from asyncio import get_running_loop
from concurrent.futures import Executor
from functools import partial, wraps
from typing import Awaitable, Callable, TypeVar, overload

from asphalt.core import Context

if sys.version_info >= (3, 10):
    from typing import Concatenate, ParamSpec
else:
    from typing_extensions import Concatenate, ParamSpec

T_Retval = TypeVar("T_Retval")
P = ParamSpec("P")


@overload
def executor(
    func_or_executor: Executor | str,
) -> Callable[
    [Callable[Concatenate[Context, P], T_Retval]],
    Callable[Concatenate[Context, P], T_Retval | Awaitable[T_Retval]],
]:
    ...


@overload
def executor(
    func_or_executor: Callable[Concatenate[Context, P], T_Retval]
) -> Callable[Concatenate[Context, P], T_Retval | Awaitable[T_Retval]]:
    ...


def executor(
    func_or_executor: Executor | str | Callable[Concatenate[Context, P], T_Retval]
) -> (
    Callable[
        [Callable[Concatenate[Context, P], T_Retval]],
        Callable[Concatenate[Context, P], T_Retval | Awaitable[T_Retval]],
    ]
    | Callable[Concatenate[Context, P], T_Retval | Awaitable[T_Retval]]
):
    """
    Decorate a function to run in an executor.

    If no executor (or ``None``) is given, the current event loop's default executor is
    used. Otherwise, the argument must be a PEP 3148 compliant thread pool executor or
    the name of an :class:`~concurrent.futures.Executor` instance.

    If a decorated callable is called in a worker thread, the executor argument is
    ignored and the wrapped function is called directly.

    Callables wrapped with this decorator must be used with ``await`` when called in the
    event loop thread.

    Example use with the default executor (``None``)::

        @executor
        def this_runs_in_threadpool(ctx):
           return do_something_cpu_intensive()

        async def request_handler(ctx):
            result = await this_runs_in_threadpool(ctx)

    With a named :class:`~concurrent.futures.Executor` resource::

        @executor('special_ops')
        def this_runs_in_threadpool(ctx):
           return do_something_cpu_intensive()

        async def request_handler(ctx):
            result = await this_runs_in_threadpool(ctx)

    :param func_or_executor: either a callable (when used as a decorator), an executor
        instance or the name of an :class:`~concurrent.futures.Executor` resource

    """

    def outer(
        func: Callable[Concatenate[Context, P], T_Retval]
    ) -> Callable[Concatenate[Context, P], T_Retval | Awaitable[T_Retval]]:
        def wrapper(
            ctx: Context, *args: P.args, **kwargs: P.kwargs
        ) -> T_Retval | Awaitable[T_Retval]:
            try:
                loop = get_running_loop()
            except RuntimeError:
                # Event loop not available -- we're in a worker thread
                return func(ctx, *args, **kwargs)

            # Resolve the executor resource name to an Executor instance
            _executor: Executor | None
            if isinstance(executor, str):
                _executor = ctx.require_resource(Executor, executor)
            else:
                _executor = executor

            callback = partial(func, ctx, *args, **kwargs)
            return loop.run_in_executor(_executor, callback)

        assert not inspect.iscoroutinefunction(
            func
        ), "Cannot wrap coroutine functions to be run in an executor"
        return wraps(func)(wrapper)

    executor: Executor | str | None = None
    if isinstance(func_or_executor, (str, Executor)):
        executor = func_or_executor
        return outer
    else:
        return outer(func_or_executor)



# isort: off
from asyncio import AbstractEventLoop

import pytest
from _pytest.capture import CaptureFixture

from asphalt.core import Context

from echo.client import ClientComponent
from echo.server import ServerComponent


def test_client_and_server(
    event_loop: AbstractEventLoop, capsys: CaptureFixture[str]
) -> None:
    async def run() -> None:
        async with Context() as ctx:
            server = ServerComponent()
            await server.start(ctx)

            client = ClientComponent("Hello!")
            await client.start(ctx)

    event_loop.create_task(run())
    with pytest.raises(SystemExit) as exc:
        event_loop.run_forever()

    assert exc.value.code == 0

    # Grab the captured output of sys.stdout and sys.stderr from the capsys fixture
    out, err = capsys.readouterr()
    assert out == "Message from client: Hello!\nServer responded: Hello!\n"



"""This is the root component for the Asphalt webnotifier tutorial."""
# isort: off
import logging
from difflib import HtmlDiff

from async_generator import aclosing
from asphalt.core import CLIApplicationComponent, Context, inject, resource
from asphalt.mailer.api import Mailer

from webnotifier.detector import ChangeDetectorComponent, Detector

logger = logging.getLogger(__name__)


class ApplicationComponent(CLIApplicationComponent):
    async def start(self, ctx: Context) -> None:
        self.add_component("detector", ChangeDetectorComponent)
        self.add_component("mailer", backend="smtp")
        await super().start(ctx)

    @inject
    async def run(
        self,
        ctx: Context,
        *,
        mailer: Mailer = resource(),
        detector: Detector = resource(),
    ) -> None:
        diff = HtmlDiff()
        async with aclosing(detector.changed.stream_events()) as stream:
            async for event in stream:
                difference = diff.make_file(
                    event.old_lines, event.new_lines, context=True
                )
                await mailer.create_and_deliver(
                    subject=f"Change detected in {event.source.url}",
                    html_body=difference,
                )
                logger.info("Sent notification email")



"""This is the client code for the Asphalt echo server tutorial."""
import sys
from asyncio import open_connection

from asphalt.core import CLIApplicationComponent, Context, run_application


class ClientComponent(CLIApplicationComponent):
    def __init__(self, message: str):
        super().__init__()
        self.message = message

    async def run(self, ctx: Context) -> None:
        reader, writer = await open_connection("localhost", 64100)
        writer.write(self.message.encode() + b"\n")
        response = await reader.readline()
        writer.close()
        print("Server responded:", response.decode().rstrip())


if __name__ == "__main__":
    component = ClientComponent(sys.argv[1])
    run_application(component)



from __future__ import annotations

__all__ = ("Component", "ContainerComponent", "CLIApplicationComponent")

import asyncio
import sys
from abc import ABCMeta, abstractmethod
from asyncio import Future
from collections import OrderedDict
from traceback import print_exception
from typing import Any, Dict, Optional, Type, Union
from warnings import warn

from typeguard import check_argument_types

from asphalt.core.context import Context
from asphalt.core.utils import PluginContainer, merge_config, qualified_name


class Component(metaclass=ABCMeta):
    """This is the base class for all Asphalt components."""

    __slots__ = ()

    @abstractmethod
    async def start(self, ctx: Context) -> None:
        """
        Perform any necessary tasks to start the services provided by this component.

        In this method, components typically use the context to:
          * add resources and/or resource factories to it
            (:meth:`~asphalt.core.context.Context.add_resource` and
            :meth:`~asphalt.core.context.Context.add_resource_factory`)
          * request resources from it asynchronously
            (:meth:`~asphalt.core.context.Context.request_resource`)

        It is advisable for Components to first add all the resources they can to the context
        before requesting any from it. This will speed up the dependency resolution and prevent
        deadlocks.

        :param ctx: the containing context for this component
        """


class ContainerComponent(Component):
    """
    A component that can contain other components.

    :param components: dictionary of component alias ⭢ component configuration dictionary

    :ivar child_components: dictionary of component alias ⭢ :class:`Component` instance (of child
        components added with :meth:`add_component`)
    :vartype child_components: Dict[str, Component]
    :ivar component_configs: dictionary of component alias ⭢ externally provided component
        configuration
    :vartype component_configs: Dict[str, Optional[Dict[str, Any]]]
    """

    __slots__ = "child_components", "component_configs"

    def __init__(self, components: Dict[str, Optional[Dict[str, Any]]] = None) -> None:
        assert check_argument_types()
        self.child_components: OrderedDict[str, Component] = OrderedDict()
        self.component_configs = components or {}

    def add_component(
        self, alias: str, type: Union[str, Type] = None, **config
    ) -> None:
        """
        Add a child component.

        This will instantiate a component class, as specified by the ``type`` argument.

        If the second argument is omitted, the value of ``alias`` is used as its value.

        The locally given configuration can be overridden by component configuration parameters
        supplied to the constructor (via the ``components`` argument).

        When configuration values are provided both as keyword arguments to this method and
        component configuration through the ``components`` constructor argument, the configurations
        are merged together using :func:`~asphalt.core.util.merge_config` in a way that the
        configuration values from the ``components`` argument override the keyword arguments to
        this method.

        :param alias: a name for the component instance, unique within this container
        :param type: entry point name or :class:`Component` subclass or a ``module:varname``
            reference to one
        :param config: keyword arguments passed to the component's constructor

        """
        assert check_argument_types()
        if not isinstance(alias, str) or not alias:
            raise TypeError("component_alias must be a nonempty string")
        if alias in self.child_components:
            raise ValueError(f'there is already a child component named "{alias}"')

        config["type"] = type or alias

        # Allow the external configuration to override the constructor arguments
        override_config = self.component_configs.get(alias) or {}
        config = merge_config(config, override_config)

        component = component_types.create_object(**config)
        self.child_components[alias] = component

    async def start(self, ctx: Context) -> None:
        """
        Create child components that have been configured but not yet created and then calls their
        :meth:`~Component.start` methods in separate tasks and waits until they have completed.

        """
        for alias in self.component_configs:
            if alias not in self.child_components:
                self.add_component(alias)

        tasks = [component.start(ctx) for component in self.child_components.values()]
        if tasks:
            await asyncio.gather(*tasks)


class CLIApplicationComponent(ContainerComponent):
    """
    Specialized subclass of :class:`.ContainerComponent` for command line tools.

    Command line tools and similar applications should use this as their root component and
    implement their main code in the :meth:`run` method.

    When all the subcomponents have been started, :meth:`run` is started as a new task.
    When the task is finished, the application will exit using the return value as its exit code.

    If :meth:`run` raises an exception, a stack trace is printed and the exit code will be set
    to 1. If the returned exit code is out of range or of the wrong data type, it is set to 1 and a
    warning is emitted.
    """

    async def start(self, ctx: Context) -> None:
        def run_complete(f: Future[int | None]) -> None:
            # If run() raised an exception, print it with a traceback and exit with code 1
            exc = f.exception()
            if exc is not None:
                print_exception(type(exc), exc, exc.__traceback__)
                sys.exit(1)

            retval = f.result()
            if isinstance(retval, int):
                if 0 <= retval <= 127:
                    sys.exit(retval)
                else:
                    warn("exit code out of range: %d" % retval)
                    sys.exit(1)
            elif retval is not None:
                warn(
                    "run() must return an integer or None, not %s"
                    % qualified_name(retval.__class__)
                )
                sys.exit(1)
            else:
                sys.exit(0)

        def start_run_task() -> None:
            task = ctx.loop.create_task(self.run(ctx))
            task.add_done_callback(run_complete)

        await super().start(ctx)
        ctx.loop.call_later(0.1, start_run_task)

    @abstractmethod
    async def run(self, ctx: Context) -> Optional[int]:
        """
        Run the business logic of the command line tool.

        Do not call this method yourself.

        :return: the application's exit code (0-127; ``None`` = 0)
        """


component_types = PluginContainer("asphalt.components", Component)



import pytest

from asphalt.core import inject, resource


@pytest.mark.asyncio
async def test_dependency_injection_posonly_argument() -> None:
    async def injected(foo: int, bar: str = resource(), /) -> None:
        pass

    pytest.raises(TypeError, inject, injected).match(
        "Cannot inject dependency to positional-only parameter 'bar'"
    )



from __future__ import annotations

import asyncio
import sys
from collections.abc import Callable
from typing import Any
from unittest.mock import Mock

import pytest

from asphalt.core.utils import (
    PluginContainer,
    callable_name,
    merge_config,
    qualified_name,
    resolve_reference,
)

if sys.version_info >= (3, 10):
    from importlib.metadata import EntryPoint
else:
    from importlib_metadata import EntryPoint


class BaseDummyPlugin:
    pass


class DummyPlugin(BaseDummyPlugin):
    def __init__(self, **kwargs) -> None:
        self.kwargs = kwargs


@pytest.mark.parametrize(
    "inputval",
    ["asphalt.core.utils:resolve_reference", resolve_reference],
    ids=["reference", "object"],
)
def test_resolve_reference(inputval: Any) -> None:
    assert resolve_reference(inputval) is resolve_reference


@pytest.mark.parametrize(
    "inputval, error_type, error_text",
    [
        (
            "x.y:foo",
            LookupError,
            "error resolving reference x.y:foo: could not import module",
        ),
        (
            "asphalt.core:foo",
            LookupError,
            "error resolving reference asphalt.core:foo: error looking up object",
        ),
    ],
    ids=["module_not_found", "object_not_found"],
)
def test_resolve_reference_error(inputval, error_type, error_text) -> None:
    exc = pytest.raises(error_type, resolve_reference, inputval)
    assert str(exc.value) == error_text


@pytest.mark.parametrize(
    "overrides",
    [
        {"a": 2, "foo": 6, "b": {"x": 5, "z": {"r": "bar", "s": [6, 7]}}},
        {"a": 2, "foo": 6, "b.x": 5, "b.z": {"r": "bar", "s": [6, 7]}},
        {"a": 2, "foo": 6, "b.x": 5, "b.z.r": "bar", "b.z.s": [6, 7]},
    ],
    ids=["nested_dicts", "part_nested", "dotted_paths"],
)
def test_merge_config(overrides) -> None:
    original = {"a": 1, "b": {"x": 2, "y": 3, "z": {"r": [1, 2], "s": [3, 4]}}}
    expected = {"a": 2, "foo": 6, "b": {"x": 5, "y": 3, "z": {"r": "bar", "s": [6, 7]}}}
    assert merge_config(original, overrides) == expected


@pytest.mark.parametrize(
    "original, overrides",
    [
        (None, {"a": 1}),
        ({"a": 1}, None),
    ],
    ids=["original_none", "override_none"],
)
def test_merge_config_none_args(original, overrides) -> None:
    assert merge_config(original, overrides) == {"a": 1}


@pytest.mark.parametrize(
    "inputval, expected",
    [
        (qualified_name, "function"),
        (asyncio.Event(), "asyncio.locks.Event"),
        (int, "int"),
    ],
    ids=["func", "instance", "builtintype"],
)
def test_qualified_name(inputval, expected) -> None:
    assert qualified_name(inputval) == expected


@pytest.mark.parametrize(
    "inputval, expected",
    [(qualified_name, "asphalt.core.utils.qualified_name"), (len, "len")],
    ids=["python", "builtin"],
)
def test_callable_name(inputval: Callable[..., Any], expected: str) -> None:
    assert callable_name(inputval) == expected


class TestPluginContainer:
    @pytest.fixture
    def container(self) -> PluginContainer:
        container = PluginContainer(
            "asphalt.core.test_plugin_container", BaseDummyPlugin
        )
        entrypoint = Mock(EntryPoint)
        entrypoint.load.configure_mock(return_value=DummyPlugin)
        container._entrypoints = {"dummy": entrypoint}
        return container

    @pytest.mark.parametrize(
        "inputvalue",
        ["dummy", "test_utils:DummyPlugin", DummyPlugin],
        ids=["entrypoint", "reference", "arbitrary_object"],
    )
    def test_resolve(self, container: PluginContainer, inputvalue) -> None:
        assert container.resolve(inputvalue) is DummyPlugin

    def test_resolve_bad_entrypoint(self, container):
        exc = pytest.raises(LookupError, container.resolve, "blah")
        assert (
            str(exc.value)
            == "no such entry point in asphalt.core.test_plugin_container: blah"
        )

    @pytest.mark.parametrize(
        "argument",
        [DummyPlugin, "dummy", "test_utils:DummyPlugin"],
        ids=["explicit_class", "entrypoint", "class_reference"],
    )
    def test_create_object(self, container: PluginContainer, argument) -> None:
        """
        Test that create_object works with all three supported ways of passing a plugin class
        reference.

        """
        component = container.create_object(argument, a=5, b=2)

        assert isinstance(component, DummyPlugin)
        assert component.kwargs == {"a": 5, "b": 2}

    def test_create_object_bad_type(self, container) -> None:
        exc = pytest.raises(TypeError, container.create_object, int)
        assert str(exc.value) == "int is not a subclass of test_utils.BaseDummyPlugin"

    def test_names(self, container: PluginContainer) -> None:
        assert container.names == ["dummy"]

    def test_all(self, container: PluginContainer) -> None:
        """
        Test that all() returns the same results before and after the entry points have been
        loaded.

        """
        assert container.all() == [DummyPlugin]
        assert container.all() == [DummyPlugin]

    def test_repr(self, container) -> None:
        assert repr(container) == (
            "PluginContainer(namespace='asphalt.core.test_plugin_container', "
            "base_class=test_utils.BaseDummyPlugin)"
        )



from __future__ import annotations

from asyncio import AbstractEventLoop
from concurrent.futures import Executor, ThreadPoolExecutor
from threading import Thread, current_thread

import pytest
import pytest_asyncio

from asphalt.core import Context
from asphalt.core.concurrent import executor


@pytest.fixture
def context() -> Context:
    return Context()


@pytest_asyncio.fixture
async def special_executor(context: Context) -> ThreadPoolExecutor:
    executor = ThreadPoolExecutor(1)
    context.add_resource(executor, "special", types=[Executor])
    context.add_teardown_callback(executor.shutdown)
    return executor


@pytest.mark.parametrize(
    "use_resource_name", [False, True], ids=["instance", "resource_name"]
)
@pytest.mark.asyncio
async def test_executor_special(
    context: Context, use_resource_name: bool, special_executor: ThreadPoolExecutor
) -> None:
    @executor("special" if use_resource_name else special_executor)
    def check_thread(ctx: Context) -> None:
        assert current_thread() is executor_thread

    async with context:
        executor_thread = special_executor.submit(current_thread).result()
        await check_thread(context)


@pytest.mark.asyncio
async def test_executor_default(
    event_loop: AbstractEventLoop, context: Context
) -> None:
    @executor
    def check_thread(ctx: Context) -> None:
        assert current_thread() is not event_loop_thread

    async with context:
        event_loop_thread = current_thread()
        await check_thread(context)


@pytest.mark.asyncio
async def test_executor_worker_thread(
    event_loop: AbstractEventLoop,
    context: Context,
    special_executor: ThreadPoolExecutor,
) -> None:
    @executor("special")
    def runs_in_special_worker(ctx: Context, worker_thread: Thread) -> str:
        assert current_thread() is worker_thread
        return "foo"

    @executor
    def runs_in_default_worker(ctx: Context) -> str:
        assert current_thread() is not event_loop_thread
        assert current_thread() is not special_executor_thread
        return runs_in_special_worker(ctx, current_thread())

    async with context:
        event_loop_thread = current_thread()
        special_executor_thread = special_executor.submit(current_thread).result()
        retval = await runs_in_default_worker(context)
        assert retval == "foo"


@pytest.mark.asyncio
async def test_executor_missing_context(
    event_loop: AbstractEventLoop, context: Context
) -> None:
    @executor("special")
    def runs_in_default_worker() -> None:
        pass

    async with context:
        with pytest.raises(TypeError) as exc:
            await runs_in_default_worker()

    exc.match(r".*?\(\) missing 1 required positional argument: 'ctx'")



from __future__ import annotations

__all__ = ("run_application",)

import asyncio
import signal
import sys
from asyncio.events import AbstractEventLoop
from concurrent.futures import ThreadPoolExecutor
from logging import INFO, Logger, basicConfig, getLogger, shutdown
from logging.config import dictConfig
from typing import Any, Dict, Optional, Union, cast

from asphalt.core.component import Component, component_types
from asphalt.core.context import Context, _current_context
from asphalt.core.utils import PluginContainer, qualified_name

policies = PluginContainer("asphalt.core.event_loop_policies")


def sigterm_handler(logger: Logger, event_loop: AbstractEventLoop) -> None:
    if event_loop.is_running():
        logger.info("Received SIGTERM")
        event_loop.stop()


def run_application(
    component: Union[Component, Dict[str, Any]],
    *,
    event_loop_policy: str | None = None,
    max_threads: int | None = None,
    logging: Union[Dict[str, Any], int, None] = INFO,
    start_timeout: Union[int, float, None] = 10,
) -> None:
    """
    Configure logging and start the given root component in the default asyncio event loop.

    Assuming the root component was started successfully, the event loop will continue running
    until the process is terminated.

    Initializes the logging system first based on the value of ``logging``:
      * If the value is a dictionary, it is passed to :func:`logging.config.dictConfig` as
        argument.
      * If the value is an integer, it is passed to :func:`logging.basicConfig` as the logging
        level.
      * If the value is ``None``, logging setup is skipped entirely.

    By default, the logging system is initialized using :func:`~logging.basicConfig` using the
    ``INFO`` logging level.

    The default executor in the event loop is replaced with a new
    :class:`~concurrent.futures.ThreadPoolExecutor` where the maximum number of threads is set to
    the value of ``max_threads`` or, if omitted, the default value of
    :class:`~concurrent.futures.ThreadPoolExecutor`.

    :param component: the root component (either a component instance or a configuration dictionary
        where the special ``type`` key is either a component class or a ``module:varname``
        reference to one)
    :param event_loop_policy: entry point name (from the ``asphalt.core.event_loop_policies``
        namespace) of an alternate event loop policy (or a module:varname reference to one)
    :param max_threads: the maximum number of worker threads in the default thread pool executor
        (the default value depends on the event loop implementation)
    :param logging: a logging configuration dictionary, :ref:`logging level <python:levels>` or
        ``None``
    :param start_timeout: seconds to wait for the root component (and its subcomponents) to start
        up before giving up (``None`` = wait forever)

    """
    # Configure the logging system
    if isinstance(logging, dict):
        dictConfig(logging)
    elif isinstance(logging, int):
        basicConfig(level=logging)

    # Inform the user whether -O or PYTHONOPTIMIZE was set when Python was launched
    logger = getLogger(__name__)
    logger.info("Running in %s mode", "development" if __debug__ else "production")

    # Switch to an alternate event loop policy if one was provided
    if event_loop_policy:
        create_policy = policies.resolve(event_loop_policy)
        policy = create_policy()
        asyncio.set_event_loop_policy(policy)
        logger.info("Switched event loop policy to %s", qualified_name(policy))

    # Assign a new default executor with the given max worker thread limit if one was provided
    event_loop = asyncio.new_event_loop()
    asyncio.set_event_loop(event_loop)
    try:
        if max_threads is not None:
            event_loop.set_default_executor(ThreadPoolExecutor(max_threads))
            logger.info(
                "Installed a new thread pool executor with max_workers=%d", max_threads
            )

        # Instantiate the root component if a dict was given
        if isinstance(component, dict):
            component = cast(Component, component_types.create_object(**component))

        logger.info("Starting application")
        context = Context()
        exception: Optional[BaseException] = None
        exit_code = 0

        # Start the root component
        token = _current_context.set(context)
        try:
            coro = asyncio.wait_for(component.start(context), start_timeout)
            event_loop.run_until_complete(coro)
        except asyncio.TimeoutError as e:
            exception = e
            logger.error("Timeout waiting for the root component to start")
            exit_code = 1
        except Exception as e:
            exception = e
            logger.exception("Error during application startup")
            exit_code = 1
        else:
            logger.info("Application started")

            # Add a signal handler to gracefully deal with SIGTERM
            try:
                event_loop.add_signal_handler(
                    signal.SIGTERM, sigterm_handler, logger, event_loop
                )
            except NotImplementedError:
                pass  # Windows does not support signals very well

            # Finally, run the event loop until the process is terminated or Ctrl+C is pressed
            try:
                event_loop.run_forever()
            except KeyboardInterrupt:
                pass
            except SystemExit as e:
                exit_code = e.code
        finally:
            _current_context.reset(token)

        # Close the root context
        logger.info("Stopping application")
        event_loop.run_until_complete(context.close(exception))

        # Shut down leftover async generators
        event_loop.run_until_complete(event_loop.shutdown_asyncgens())
    finally:
        # Finally, close the event loop itself
        event_loop.close()
        asyncio.set_event_loop(None)
        logger.info("Application stopped")

    # Shut down the logging system
    shutdown()

    if exit_code:
        sys.exit(exit_code)



__all__ = (
    "CLIApplicationComponent",
    "Component",
    "ContainerComponent",
    "Context",
    "ResourceConflict",
    "ResourceEvent",
    "ResourceNotFound",
    "TeardownError",
    "context_teardown",
    "current_context",
    "get_resource",
    "get_resources",
    "require_resource",
    "NoCurrentContext",
    "Dependency",
    "inject",
    "resource",
    "executor",
    "Event",
    "Signal",
    "stream_events",
    "wait_event",
    "run_application",
    "PluginContainer",
    "callable_name",
    "merge_config",
    "qualified_name",
    "resolve_reference",
)

from .component import CLIApplicationComponent, Component, ContainerComponent
from .context import (
    Context,
    Dependency,
    NoCurrentContext,
    ResourceConflict,
    ResourceEvent,
    ResourceNotFound,
    TeardownError,
    context_teardown,
    current_context,
    executor,
    get_resource,
    get_resources,
    inject,
    require_resource,
    resource,
)
from .event import Event, Signal, stream_events, wait_event
from .runner import run_application
from .utils import (
    PluginContainer,
    callable_name,
    merge_config,
    qualified_name,
    resolve_reference,
)



from __future__ import annotations

import asyncio
import sys
from collections.abc import Callable
from concurrent.futures import Executor, ThreadPoolExecutor
from inspect import isawaitable
from itertools import count
from threading import Thread, current_thread
from typing import AsyncGenerator, AsyncIterator, Dict, NoReturn, Optional, Tuple, Union
from unittest.mock import patch

import pytest
import pytest_asyncio
from async_generator import yield_

from asphalt.core import (
    Context,
    Dependency,
    NoCurrentContext,
    ResourceConflict,
    ResourceNotFound,
    TeardownError,
    callable_name,
    context_teardown,
    current_context,
    executor,
    get_resource,
    inject,
    resource,
)
from asphalt.core.context import ResourceContainer, require_resource


@pytest.fixture
def context() -> Context:
    return Context()


@pytest_asyncio.fixture
async def special_executor(context: Context) -> AsyncIterator[ThreadPoolExecutor]:
    executor = ThreadPoolExecutor(1)
    context.add_resource(executor, "special", types=[Executor])
    yield executor
    executor.shutdown()


class TestResourceContainer:
    @pytest.mark.parametrize("thread", [False, True], ids=["eventloop", "worker"])
    @pytest.mark.parametrize(
        "context_attr", [None, "attrname"], ids=["no_attr", "has_attr"]
    )
    @pytest.mark.asyncio
    async def test_generate_value(self, thread: bool, context_attr: str | None) -> None:
        container = ResourceContainer(
            lambda ctx: "foo", (str,), "default", context_attr, True
        )
        context = Context()
        if thread:
            value = await context.call_in_executor(container.generate_value, context)
        else:
            value = container.generate_value(context)

        assert value == "foo"
        assert context.get_resource(str) == "foo"
        if context_attr:
            assert getattr(context, context_attr) == "foo"

    def test_repr(self) -> None:
        container = ResourceContainer("foo", (str,), "default", "attrname", False)
        assert repr(container) == (
            "ResourceContainer(value='foo', types=[str], name='default', "
            "context_attr='attrname')"
        )

    def test_repr_factory(self) -> None:
        container = ResourceContainer(
            lambda ctx: "foo", (str,), "default", "attrname", True
        )
        assert repr(container) == (
            "ResourceContainer(factory=test_context.TestResourceContainer"
            ".test_repr_factory.<locals>.<lambda>, types=[str], name='default', "
            "context_attr='attrname')"
        )


class TestContext:
    @pytest.mark.asyncio
    async def test_parent(self) -> None:
        """Test that the parent property points to the parent context instance, if any."""
        async with Context() as parent:
            async with Context() as child:
                assert parent.parent is None
                assert child.parent is parent

    @pytest.mark.parametrize(
        "exception", [None, Exception("foo")], ids=["noexception", "exception"]
    )
    @pytest.mark.asyncio
    async def test_close(self, context: Context, exception: Exception | None) -> None:
        """
        Test that teardown callbacks are called in reverse order when a context is closed.

        """

        def callback(exception=None):
            called_functions.append((callback, exception))

        async def async_callback(exception=None):
            called_functions.append((async_callback, exception))

        called_functions: list[tuple[Callable, BaseException | None]] = []
        context.add_teardown_callback(callback, pass_exception=True)
        context.add_teardown_callback(async_callback, pass_exception=True)
        await context.close(exception)

        assert called_functions == [(async_callback, exception), (callback, exception)]

    @pytest.mark.asyncio
    async def test_close_while_running_teardown(self, context: Context) -> None:
        """
        Test that trying to close the context from a teardown callback raises a
        RuntimeError.
        """

        async def try_close_context() -> None:
            with pytest.raises(RuntimeError, match="this context is already closing"):
                await context.close()

        context.add_teardown_callback(try_close_context)
        await context.close()

    @pytest.mark.asyncio
    async def test_teardown_callback_exception(self, context: Context) -> None:
        """
        Test that all callbacks are called even when some teardown callbacks raise
        exceptions, and that a TeardownError is raised in such a case, containing the
        exception objects.

        """

        def callback1() -> None:
            items.append(1)

        def callback2() -> NoReturn:
            raise Exception("foo")

        context.add_teardown_callback(callback1)
        context.add_teardown_callback(callback2)
        context.add_teardown_callback(callback1)
        context.add_teardown_callback(callback2)
        items: list[int] = []
        with pytest.raises(TeardownError) as exc:
            await context.close()

        assert "foo" in str(exc.value)
        assert items == [1, 1]
        assert len(exc.value.exceptions) == 2

    @pytest.mark.asyncio
    async def test_close_closed(self, context: Context) -> None:
        """Test that closing an already closed context raises a RuntimeError."""
        assert not context.closed
        await context.close()
        assert context.closed

        with pytest.raises(RuntimeError) as exc:
            await context.close()

        exc.match("this context has already been closed")

    def test_contextmanager_exception(self, context, event_loop):
        close_future = event_loop.create_future()
        close_future.set_result(None)
        exception = Exception("foo")
        with patch.object(context, "close", return_value=close_future):
            with pytest.raises(Exception) as exc, pytest.deprecated_call():
                with context:
                    raise exception

        # close.assert_called_once_with(exception)
        assert exc.value is exception

    @pytest.mark.asyncio
    async def test_async_contextmanager_exception(self, event_loop, context):
        """Test that "async with context:" calls close() with the exception raised in the block."""
        close_future = event_loop.create_future()
        close_future.set_result(None)
        exception = Exception("foo")
        with patch.object(context, "close", return_value=close_future) as close:
            with pytest.raises(Exception) as exc:
                async with context:
                    raise exception

        close.assert_called_once_with(exception)
        assert exc.value is exception

    @pytest.mark.parametrize("types", [int, (int,), ()], ids=["type", "tuple", "empty"])
    @pytest.mark.asyncio
    async def test_add_resource(self, context, event_loop, types):
        """Test that a resource is properly added in the context and listeners are notified."""
        event_loop.call_soon(context.add_resource, 6, "foo", None, types)
        event = await context.resource_added.wait_event()

        assert event.resource_types == (int,)
        assert event.resource_name == "foo"
        assert not event.is_factory
        assert context.get_resource(int, "foo") == 6

    @pytest.mark.asyncio
    async def test_add_resource_name_conflict(self, context: Context) -> None:
        """Test that adding a resource won't replace any existing resources."""
        context.add_resource(5, "foo")
        with pytest.raises(ResourceConflict) as exc:
            context.add_resource(4, "foo")

        exc.match(
            "this context already contains a resource of type int using the name 'foo'"
        )

    @pytest.mark.asyncio
    async def test_add_resource_none_value(self, context: Context) -> None:
        """Test that None is not accepted as a resource value."""
        exc = pytest.raises(ValueError, context.add_resource, None)
        exc.match('"value" must not be None')

    @pytest.mark.asyncio
    async def test_add_resource_context_attr(self, context: Context) -> None:
        """Test that when resources are added, they are also set as properties of the context."""
        with pytest.deprecated_call():
            context.add_resource(1, context_attr="foo")

        assert context.foo == 1

    def test_add_resource_context_attr_conflict(self, context: Context) -> None:
        """
        Test that the context won't allow adding a resource with an attribute name that conflicts
        with an existing attribute.

        """
        context.a = 2
        with pytest.raises(ResourceConflict) as exc, pytest.deprecated_call():
            context.add_resource(2, context_attr="a")

        exc.match("this context already has an attribute 'a'")
        assert context.get_resource(int) is None

    @pytest.mark.asyncio
    async def test_add_resource_type_conflict(self, context: Context) -> None:
        context.add_resource(5)
        with pytest.raises(ResourceConflict) as exc:
            context.add_resource(6)

        exc.match(
            "this context already contains a resource of type int using the name 'default'"
        )

    @pytest.mark.parametrize(
        "name", ["a.b", "a:b", "a b"], ids=["dot", "colon", "space"]
    )
    @pytest.mark.asyncio
    async def test_add_resource_bad_name(self, context, name):
        with pytest.raises(ValueError) as exc:
            context.add_resource(1, name)

        exc.match(
            '"name" must be a nonempty string consisting only of alphanumeric characters '
            "and underscores"
        )

    @pytest.mark.asyncio
    async def test_add_resource_parametrized_generic_type(
        self, context: Context
    ) -> None:
        resource = {"a": 1}
        resource_type = Dict[str, int]
        context.add_resource(resource, types=[resource_type])
        assert context.require_resource(resource_type) is resource
        assert context.get_resource(resource_type) is resource
        assert await context.request_resource(resource_type) is resource
        assert context.get_resource(Dict) is None
        assert context.get_resource(dict) is None

    @pytest.mark.asyncio
    async def test_add_resource_factory(self, context: Context) -> None:
        """Test that resources factory callbacks are only called once for each context."""

        def factory(ctx):
            assert ctx is context
            return next(counter)

        counter = count(1)
        with pytest.deprecated_call():
            context.add_resource_factory(factory, int, context_attr="foo")

        assert context.foo == 1
        assert context.foo == 1
        assert context.__dict__["foo"] == 1

    @pytest.mark.asyncio
    async def test_add_resource_factory_parametrized_generic_type(
        self, context: Context
    ) -> None:
        resource = {"a": 1}
        resource_type = Dict[str, int]
        context.add_resource_factory(lambda ctx: resource, types=[resource_type])
        assert context.require_resource(resource_type) is resource
        assert context.get_resource(resource_type) is resource
        assert await context.request_resource(resource_type) is resource
        assert context.get_resource(Dict) is None
        assert context.get_resource(dict) is None

    @pytest.mark.parametrize(
        "name", ["a.b", "a:b", "a b"], ids=["dot", "colon", "space"]
    )
    @pytest.mark.asyncio
    async def test_add_resource_factory_bad_name(self, context, name):
        with pytest.raises(ValueError) as exc:
            context.add_resource_factory(lambda ctx: 1, int, name)

        exc.match(
            '"name" must be a nonempty string consisting only of alphanumeric characters '
            "and underscores"
        )

    @pytest.mark.asyncio
    async def test_add_resource_factory_coroutine_callback(
        self, context: Context
    ) -> None:
        async def factory(ctx):
            return 1

        with pytest.raises(TypeError) as exc:
            context.add_resource_factory(factory, int)

        exc.match('"factory_callback" must not be a coroutine function')

    @pytest.mark.asyncio
    async def test_add_resource_factory_empty_types(self, context: Context) -> None:
        with pytest.raises(ValueError) as exc:
            context.add_resource_factory(lambda ctx: 1, ())

        exc.match("no resource types were specified")

    @pytest.mark.asyncio
    async def test_add_resource_factory_context_attr_conflict(
        self, context: Context
    ) -> None:
        with pytest.deprecated_call():
            context.add_resource_factory(lambda ctx: None, str, context_attr="foo")

        with pytest.raises(ResourceConflict) as exc, pytest.deprecated_call():
            await context.add_resource_factory(
                lambda ctx: None, str, context_attr="foo"
            )

        exc.match(
            "this context already contains a resource factory for the context attribute 'foo'"
        )

    @pytest.mark.asyncio
    async def test_add_resource_factory_type_conflict(self, context: Context) -> None:
        context.add_resource_factory(lambda ctx: None, (str, int))
        with pytest.raises(ResourceConflict) as exc:
            await context.add_resource_factory(lambda ctx: None, int)

        exc.match("this context already contains a resource factory for the type int")

    @pytest.mark.asyncio
    async def test_add_resource_factory_no_inherit(self, context: Context) -> None:
        """
        Test that a subcontext gets its own version of a factory-generated resource even if a
        parent context has one already.

        """
        with pytest.deprecated_call():
            context.add_resource_factory(id, int, context_attr="foo")

        async with context, Context() as subcontext:
            assert context.foo == id(context)
            assert subcontext.foo == id(subcontext)

    @pytest.mark.asyncio
    async def test_add_resource_return_type_single(self, context: Context) -> None:
        def factory(ctx: Context) -> str:
            return "foo"

        async with context:
            context.add_resource_factory(factory)
            assert context.require_resource(str) == "foo"

    @pytest.mark.asyncio
    async def test_add_resource_return_type_union(self, context: Context) -> None:
        def factory(ctx: Context) -> Union[int, float]:
            return 5

        async with context:
            context.add_resource_factory(factory)
            assert context.require_resource(int) == 5
            assert context.require_resource(float) == 5

    @pytest.mark.skipif(sys.version_info < (3, 10), reason="Requires Python 3.10+")
    @pytest.mark.asyncio
    async def test_add_resource_return_type_uniontype(self, context: Context) -> None:
        def factory(ctx: Context) -> int | float:
            return 5

        async with context:
            context.add_resource_factory(factory)
            assert context.require_resource(int) == 5
            assert context.require_resource(float) == 5

    @pytest.mark.asyncio
    async def test_add_resource_return_type_optional(self, context: Context) -> None:
        def factory(ctx: Context) -> Optional[str]:
            return "foo"

        async with context:
            context.add_resource_factory(factory)
            assert context.require_resource(str) == "foo"

    @pytest.mark.asyncio
    async def test_getattr_attribute_error(self, context: Context) -> None:
        async with context, Context() as child_context:
            pytest.raises(AttributeError, getattr, child_context, "foo").match(
                "no such context variable: foo"
            )

    @pytest.mark.asyncio
    async def test_getattr_parent(self, context: Context) -> None:
        """
        Test that accessing a nonexistent attribute on a context retrieves the value from parent.

        """
        async with context, Context() as child_context:
            context.a = 2
            assert child_context.a == 2

    @pytest.mark.asyncio
    async def test_get_resources(self, context: Context) -> None:
        context.add_resource(9, "foo")
        context.add_resource_factory(lambda ctx: len(ctx.context_chain), int, "bar")
        context.require_resource(int, "bar")
        async with context, Context() as subctx:
            subctx.add_resource(4, "foo")
            assert subctx.get_resources(int) == {1, 4}

    @pytest.mark.asyncio
    async def test_require_resource(self, context: Context) -> None:
        context.add_resource(1)
        assert context.require_resource(int) == 1

    def test_require_resource_not_found(self, context: Context) -> None:
        """Test that ResourceNotFound is raised when a required resource is not found."""
        exc = pytest.raises(ResourceNotFound, context.require_resource, int, "foo")
        exc.match("no matching resource was found for type=int name='foo'")
        assert exc.value.type == int
        assert exc.value.name == "foo"

    @pytest.mark.asyncio
    async def test_request_resource_parent_add(self, context, event_loop):
        """
        Test that adding a resource to the parent context will satisfy a resource request in a
        child context.

        """
        async with context, Context() as child_context:
            task = event_loop.create_task(child_context.request_resource(int))
            event_loop.call_soon(context.add_resource, 6)
            resource = await task
            assert resource == 6

    @pytest.mark.asyncio
    async def test_request_resource_factory_context_attr(
        self, context: Context
    ) -> None:
        """Test that requesting a factory-generated resource also sets the context variable."""
        with pytest.deprecated_call():
            context.add_resource_factory(lambda ctx: 6, int, context_attr="foo")

        await context.request_resource(int)
        assert context.__dict__["foo"] == 6

    @pytest.mark.asyncio
    async def test_call_async_plain(self, context: Context) -> None:
        def runs_in_event_loop(worker_thread: Thread, x: int, y: int) -> int:
            assert current_thread() is not worker_thread
            return x + y

        def runs_in_worker_thread() -> int:
            worker_thread = current_thread()
            return context.call_async(runs_in_event_loop, worker_thread, 1, y=2)

        assert await context.call_in_executor(runs_in_worker_thread) == 3

    @pytest.mark.asyncio
    async def test_call_async_coroutine(self, context: Context) -> None:
        async def runs_in_event_loop(worker_thread, x, y):
            assert current_thread() is not worker_thread
            await asyncio.sleep(0.1)
            return x + y

        def runs_in_worker_thread() -> int:
            worker_thread = current_thread()
            return context.call_async(runs_in_event_loop, worker_thread, 1, y=2)

        assert await context.call_in_executor(runs_in_worker_thread) == 3

    @pytest.mark.asyncio
    async def test_call_async_exception(self, context: Context) -> None:
        def runs_in_event_loop() -> NoReturn:
            raise ValueError("foo")

        with pytest.raises(ValueError) as exc:
            await context.call_in_executor(context.call_async, runs_in_event_loop)

        assert exc.match("foo")

    @pytest.mark.asyncio
    async def test_call_in_executor(self, context: Context) -> None:
        """Test that call_in_executor actually runs the target in a worker thread."""
        worker_thread = await context.call_in_executor(current_thread)
        assert worker_thread is not current_thread()

    @pytest.mark.parametrize(
        "use_resource_name", [True, False], ids=["direct", "resource"]
    )
    @pytest.mark.asyncio
    async def test_call_in_executor_explicit(self, context, use_resource_name):
        executor = ThreadPoolExecutor(1)
        context.add_resource(executor, types=[Executor])
        context.add_teardown_callback(executor.shutdown)
        executor_arg = "default" if use_resource_name else executor
        worker_thread = await context.call_in_executor(
            current_thread, executor=executor_arg
        )
        assert worker_thread is not current_thread()

    @pytest.mark.asyncio
    async def test_call_in_executor_context_preserved(self, context: Context) -> None:
        """
        Test that call_in_executor runs the callable in a copy of the current (PEP 567)
        context.
        """

        async with Context() as ctx:
            assert await context.call_in_executor(current_context) is ctx

    @pytest.mark.asyncio
    async def test_threadpool(self, context: Context) -> None:
        event_loop_thread = current_thread()
        async with context.threadpool():
            assert current_thread() is not event_loop_thread

    @pytest.mark.asyncio
    async def test_threadpool_named_executor(self, context, special_executor):
        special_executor_thread = special_executor.submit(current_thread).result()
        async with context.threadpool("special"):
            assert current_thread() is special_executor_thread


class TestExecutor:
    @pytest.mark.asyncio
    async def test_no_arguments(self, context: Context) -> None:
        @executor
        def runs_in_default_worker() -> None:
            assert current_thread() is not event_loop_thread

        event_loop_thread = current_thread()
        await runs_in_default_worker()

    @pytest.mark.asyncio
    async def test_named_executor(self, context, special_executor):
        @executor("special")
        def runs_in_default_worker(ctx: Context) -> None:
            assert current_thread() is special_executor_thread

        special_executor_thread = special_executor.submit(current_thread).result()
        await runs_in_default_worker(context)

    @pytest.mark.asyncio
    async def test_executor_missing_context(self, event_loop, context):
        @executor("special")
        def runs_in_default_worker() -> None:
            pass

        with pytest.raises(RuntimeError) as exc:
            await runs_in_default_worker()

        exc.match(
            r"the first positional argument to %s\(\) has to be a Context instance"
            % callable_name(runs_in_default_worker)
        )


class TestContextTeardown:
    @pytest.mark.parametrize(
        "expected_exc", [None, Exception("foo")], ids=["no_exception", "exception"]
    )
    @pytest.mark.asyncio
    async def test_function(self, expected_exc: Exception | None) -> None:
        phase = received_exception = None

        @context_teardown
        async def start(ctx: Context) -> AsyncIterator[None]:
            nonlocal phase, received_exception
            phase = "started"
            exc = yield
            phase = "finished"
            received_exception = exc

        context = Context()
        await start(context)
        assert phase == "started"

        await context.close(expected_exc)
        assert phase == "finished"
        assert received_exception == expected_exc

    @pytest.mark.parametrize(
        "expected_exc", [None, Exception("foo")], ids=["no_exception", "exception"]
    )
    @pytest.mark.asyncio
    async def test_method(self, expected_exc: Exception | None) -> None:
        phase = received_exception = None

        class SomeComponent:
            @context_teardown
            async def start(self, ctx: Context) -> AsyncIterator[None]:
                nonlocal phase, received_exception
                phase = "started"
                exc = yield
                phase = "finished"
                received_exception = exc

        context = Context()
        await SomeComponent().start(context)
        assert phase == "started"

        await context.close(expected_exc)
        assert phase == "finished"
        assert received_exception == expected_exc

    def test_plain_function(self) -> None:
        def start(ctx) -> None:
            pass

        pytest.raises(TypeError, context_teardown, start).match(
            " must be an async generator function"
        )

    @pytest.mark.asyncio
    async def test_bad_args(self) -> None:
        with pytest.deprecated_call():

            @context_teardown
            async def start(ctx: Context) -> None:
                pass

        with pytest.raises(RuntimeError) as exc:
            await start(None)

        exc.match(
            r"the first positional argument to %s\(\) has to be a Context instance"
            % callable_name(start)
        )

    @pytest.mark.asyncio
    async def test_exception(self) -> None:
        @context_teardown
        async def start(ctx: Context) -> AsyncIterator[None]:
            raise Exception("dummy error")
            yield

        context = Context()
        with pytest.raises(Exception) as exc_info:
            await start(context)

        exc_info.match("dummy error")

    @pytest.mark.asyncio
    async def test_missing_yield(self) -> None:
        with pytest.deprecated_call():

            @context_teardown
            async def start(ctx: Context) -> None:
                pass

        await start(Context())

    @pytest.mark.asyncio
    async def test_py35_generator(self) -> None:
        with pytest.deprecated_call():

            @context_teardown
            async def start(ctx: Context) -> None:
                await yield_()

        await start(Context())

    @pytest.mark.parametrize(
        "resource_func",
        [
            pytest.param(Context.get_resource, id="get_resource"),
            pytest.param(Context.require_resource, id="require_resource"),
            pytest.param(Context.request_resource, id="request_resource"),
        ],
    )
    @pytest.mark.asyncio
    async def test_get_resource_at_teardown(self, resource_func) -> None:
        resource: str

        async def teardown_callback() -> None:
            nonlocal resource
            resource = resource_func(ctx, str)
            if isawaitable(resource):
                resource = await resource

        async with Context() as ctx:
            ctx.add_resource("blah")
            ctx.add_teardown_callback(teardown_callback)

        assert resource == "blah"

    @pytest.mark.parametrize(
        "resource_func",
        [
            pytest.param(Context.get_resource, id="get_resource"),
            pytest.param(Context.require_resource, id="require_resource"),
            pytest.param(Context.request_resource, id="request_resource"),
        ],
    )
    @pytest.mark.asyncio
    async def test_generate_resource_at_teardown(self, resource_func) -> None:
        resource: str

        async def teardown_callback() -> None:
            nonlocal resource
            resource = resource_func(ctx, str)
            if isawaitable(resource):
                resource = await resource

        async with Context() as ctx:
            ctx.add_resource_factory(lambda context: "blah", [str])
            ctx.add_teardown_callback(teardown_callback)

        assert resource == "blah"


class TestContextFinisher:
    @pytest.mark.parametrize(
        "expected_exc", [None, Exception("foo")], ids=["no_exception", "exception"]
    )
    @pytest.mark.asyncio
    async def test_context_teardown(self, expected_exc: Exception | None) -> None:
        phase = received_exception = None

        @context_teardown
        async def start(ctx: Context) -> AsyncIterator[None]:
            nonlocal phase, received_exception
            phase = "started"
            exc = yield
            phase = "finished"
            received_exception = exc

        context = Context()
        await start(context)
        assert phase == "started"

        await context.close(expected_exc)
        assert phase == "finished"
        assert received_exception == expected_exc


@pytest.mark.asyncio
async def test_current_context() -> None:
    pytest.raises(NoCurrentContext, current_context)

    async with Context() as parent_ctx:
        assert current_context() is parent_ctx
        async with Context() as child_ctx:
            assert current_context() is child_ctx

        assert current_context() is parent_ctx

    pytest.raises(NoCurrentContext, current_context)


@pytest.mark.asyncio
async def test_get_resource() -> None:
    async with Context() as ctx:
        ctx.add_resource("foo")
        assert get_resource(str) == "foo"
        assert get_resource(int) is None


@pytest.mark.asyncio
async def test_require_resource() -> None:
    async with Context() as ctx:
        ctx.add_resource("foo")
        assert require_resource(str) == "foo"
        pytest.raises(ResourceNotFound, require_resource, int)


def test_explicit_parent_deprecation() -> None:
    parent_ctx = Context()
    pytest.warns(DeprecationWarning, Context, parent_ctx)


@pytest.mark.asyncio
async def test_context_stack_corruption(event_loop):
    async def generator() -> AsyncGenerator:
        async with Context():
            yield

    gen = generator()
    await event_loop.create_task(gen.asend(None))
    async with Context() as ctx:
        with pytest.warns(
            UserWarning, match="Potential context stack corruption detected"
        ):
            try:
                await event_loop.create_task(gen.asend(None))
            except StopAsyncIteration:
                pass

        assert current_context() is ctx

    pytest.raises(NoCurrentContext, current_context)


class TestDependencyInjection:
    @pytest.mark.asyncio
    async def test_static_resources(self) -> None:
        @inject
        async def injected(
            foo: int, bar: str = resource(), *, baz: str = resource("alt")
        ) -> Tuple[int, str, str]:
            return foo, bar, baz

        async with Context() as ctx:
            ctx.add_resource("bar_test")
            ctx.add_resource("baz_test", "alt")
            foo, bar, baz = await injected(2)

        assert foo == 2
        assert bar == "bar_test"
        assert baz == "baz_test"

    @pytest.mark.asyncio
    async def test_sync_injection(self) -> None:
        @inject
        def injected(
            foo: int, bar: str = resource(), *, baz: str = resource("alt")
        ) -> Tuple[int, str, str]:
            return foo, bar, baz

        async with Context() as ctx:
            ctx.add_resource("bar_test")
            ctx.add_resource("baz_test", "alt")
            foo, bar, baz = injected(2)

        assert foo == 2
        assert bar == "bar_test"
        assert baz == "baz_test"

    @pytest.mark.asyncio
    async def test_missing_annotation(self) -> None:
        async def injected(
            foo: int, bar: str = resource(), *, baz=resource("alt")
        ) -> None:
            pass

        pytest.raises(TypeError, inject, injected).match(
            f"Dependency for parameter 'baz' of function "
            f"'{__name__}.{self.__class__.__name__}.test_missing_annotation.<locals>"
            f".injected' is missing the type annotation"
        )

    @pytest.mark.asyncio
    async def test_missing_resource(self) -> None:
        @inject
        async def injected(foo: int, bar: str = resource()) -> None:
            pass

        with pytest.raises(ResourceNotFound) as exc:
            async with Context():
                await injected(2)

        exc.match("no matching resource was found for type=str name='default'")

    @pytest.mark.parametrize(
        "annotation",
        [
            pytest.param(Optional[str], id="optional"),
            # pytest.param(Union[str, int, None], id="union"),
            pytest.param(
                "str | None",
                id="uniontype.10",
                marks=[
                    pytest.mark.skipif(
                        sys.version_info < (3, 10), reason="Requires Python 3.10+"
                    )
                ],
            ),
        ],
    )
    @pytest.mark.parametrize(
        "sync",
        [
            pytest.param(True, id="sync"),
            pytest.param(False, id="async"),
        ],
    )
    @pytest.mark.asyncio
    async def test_inject_optional_resource_async(
        self, annotation: type, sync: bool
    ) -> None:
        if sync:

            @inject
            def injected(
                res: annotation = resource(),  # type: ignore[valid-type]
            ) -> annotation:  # type: ignore[valid-type]
                return res

        else:

            @inject
            async def injected(
                res: annotation = resource(),  # type: ignore[valid-type]
            ) -> annotation:  # type: ignore[valid-type]
                return res

        async with Context() as ctx:
            retval = injected() if sync else (await injected())
            assert retval is None
            ctx.add_resource("hello")
            retval = injected() if sync else (await injected())
            assert retval == "hello"

    def test_resource_function_not_called(self) -> None:
        async def injected(foo: int, bar: str = resource) -> None:
            pass

        with pytest.raises(TypeError) as exc:
            inject(injected)

        exc.match(
            f"Default value for parameter 'bar' of function "
            f"{__name__}.{self.__class__.__name__}.test_resource_function_not_called"
            f".<locals>.injected was the 'resource' function – did you forget to add "
            f"the parentheses at the end\\?"
        )

    def test_missing_inject(self) -> None:
        def injected(foo: int, bar: str = resource()) -> None:
            bar.lower()

        with pytest.raises(AttributeError) as exc:
            injected(1)

        exc.match(
            r"Attempted to access an attribute in a resource\(\) marker – did you "
            r"forget to add the @inject decorator\?"
        )

    def test_no_resources_declared(self) -> None:
        def injected(foo: int) -> None:
            pass

        match = (
            f"{__name__}.{self.__class__.__name__}.test_no_resources_declared.<locals>"
            f".injected does not have any injectable resources declared"
        )
        with pytest.warns(UserWarning, match=match):
            func = inject(injected)

        assert func is injected


def test_dependency_deprecated() -> None:
    with pytest.deprecated_call():

        async def foo(res: str = Dependency()) -> None:
            pass



from __future__ import annotations

__all__ = (
    "resolve_reference",
    "qualified_name",
    "callable_name",
    "merge_config",
    "PluginContainer",
)

import sys
from importlib import import_module
from inspect import isclass
from typing import Any, Callable, Dict, List, Optional, Type, TypeVar, Union, overload

from typeguard import check_argument_types

if sys.version_info >= (3, 10):
    from importlib.metadata import EntryPoint, entry_points
else:
    from importlib_metadata import EntryPoint, entry_points

T_Object = TypeVar("T_Object")


def resolve_reference(ref: object) -> Any:
    """
    Return the object pointed to by ``ref``.

    If ``ref`` is not a string or does not contain ``:``, it is returned as is.

    References must be in the form  <modulename>:<varname> where <modulename> is the fully
    qualified module name and varname is the path to the variable inside that module.

    For example, "concurrent.futures:Future" would give you the
    :class:`~concurrent.futures.Future` class.

    :raises LookupError: if the reference could not be resolved

    """
    if not isinstance(ref, str) or ":" not in ref:
        return ref

    modulename, rest = ref.split(":", 1)
    try:
        obj = import_module(modulename)
    except ImportError as e:
        raise LookupError(
            f"error resolving reference {ref}: could not import module"
        ) from e

    try:
        for name in rest.split("."):
            obj = getattr(obj, name)

        return obj
    except AttributeError:
        raise LookupError(f"error resolving reference {ref}: error looking up object")


def qualified_name(obj: object) -> str:
    """
    Return the qualified name (e.g. package.module.Type) for the given object.

    If ``obj`` is not a class, the returned name will match its type instead.

    """
    cls = obj if isclass(obj) else type(obj)
    if cls.__module__ == "builtins":
        return cls.__name__
    else:
        return f"{cls.__module__}.{cls.__qualname__}"


def callable_name(func: Callable[..., Any]) -> str:
    """Return the qualified name (e.g. package.module.func) for the given callable."""
    if func.__module__ == "builtins":
        return func.__name__
    else:
        return f"{func.__module__}.{func.__qualname__}"


def merge_config(
    original: Optional[Dict[str, Any]], overrides: Optional[Dict[str, Any]]
) -> Dict[str, Any]:
    """
    Return a copy of the ``original`` configuration dictionary, with overrides from ``overrides``
    applied.

    This similar to what :meth:`dict.update` does, but when a dictionary is about to be
    replaced with another dictionary, it instead merges the contents.

    If a key in ``overrides`` is a dotted path (ie. ``foo.bar.baz: value``), it is assumed to be a
    shorthand for ``foo: {bar: {baz: value}}``.

    :param original: a configuration dictionary (or ``None``)
    :param overrides: a dictionary containing overriding values to the configuration (or ``None``)
    :return: the merge result

    """
    assert check_argument_types()
    copied = original.copy() if original else {}
    if overrides:
        for key, value in overrides.items():
            if "." in key:
                key, rest = key.split(".", 1)
                value = {rest: value}

            orig_value = copied.get(key)
            if isinstance(orig_value, dict) and isinstance(value, dict):
                copied[key] = merge_config(orig_value, value)
            else:
                copied[key] = value

    return copied


class PluginContainer:
    """
    A convenience class for loading and instantiating plugins through the use of entry points.

    :param namespace: a setuptools entry points namespace
    :param base_class: the base class for plugins of this type (or ``None`` if the entry points
        don't point to classes)
    """

    __slots__ = "namespace", "base_class", "_entrypoints"

    def __init__(self, namespace: str, base_class: type = None) -> None:
        self.namespace = namespace
        self.base_class = base_class
        group = entry_points().select(group=namespace)  # type: ignore[attr-defined]
        self._entrypoints = {ep.name: ep for ep in group}

    @overload
    def resolve(self, obj: str) -> Any:
        pass

    @overload
    def resolve(self, obj: T_Object) -> T_Object:
        pass

    def resolve(self, obj: Any) -> Any:
        """
        Resolve a reference to an entry point or a variable in a module.

        If ``obj`` is a ``module:varname`` reference to an object, :func:`resolve_reference` is
        used to resolve it. If it is a string of any other kind, the named entry point is loaded
        from this container's namespace. Otherwise, ``obj`` is returned as is.

        :param obj: an entry point identifier, an object reference or an arbitrary object
        :return: the loaded entry point, resolved object or the unchanged input value
        :raises LookupError: if ``obj`` was a string but the named entry point was not found

        """
        if not isinstance(obj, str):
            return obj
        if ":" in obj:
            return resolve_reference(obj)

        value = self._entrypoints.get(obj)
        if value is None:
            raise LookupError(f"no such entry point in {self.namespace}: {obj}")

        if isinstance(value, EntryPoint):
            value = self._entrypoints[obj] = value.load()

        return value

    def create_object(self, type: Union[Type, str], **constructor_kwargs) -> Any:
        """
        Instantiate a plugin.

        The entry points in this namespace must point to subclasses of the ``base_class`` parameter
        passed to this container.

        :param type: an entry point identifier, a ``module:varname`` reference to a class, or an
            actual class object
        :param constructor_kwargs: keyword arguments passed to the constructor of the plugin class
        :return: the plugin instance

        """
        assert check_argument_types()
        assert self.base_class, "base class has not been defined"
        plugin_class = self.resolve(type)
        if not isclass(plugin_class) or not issubclass(plugin_class, self.base_class):
            raise TypeError(
                "{} is not a subclass of {}".format(
                    qualified_name(plugin_class), qualified_name(self.base_class)
                )
            )

        return plugin_class(**constructor_kwargs)

    @property
    def names(self) -> List[str]:
        """Return names of all entry points in this namespace."""
        return list(self._entrypoints)

    def all(self) -> List[Any]:
        """
        Load all entry points (if not already loaded) in this namespace and return the resulting
        objects as a list.

        """
        values = []
        for name, value in self._entrypoints.items():
            if isinstance(value, EntryPoint):
                value = self._entrypoints[name] = value.load()

            values.append(value)

        return values

    def __repr__(self):
        return (
            "{0.__class__.__name__}(namespace={0.namespace!r}, base_class={1})".format(
                self, qualified_name(self.base_class)
            )
        )



from __future__ import annotations

from pathlib import Path
from unittest.mock import patch

import pytest
from _pytest.monkeypatch import MonkeyPatch
from click.testing import CliRunner

from asphalt.core import Component, Context, cli


class DummyComponent(Component):
    def __init__(self, dummyval1=None, dummyval2=None):
        self.dummyval1 = dummyval1
        self.dummyval2 = dummyval2

    def start(self, ctx: Context) -> None:
        pass


@pytest.fixture
def runner() -> CliRunner:
    return CliRunner()


@pytest.mark.parametrize("loop", [None, "uvloop"], ids=["default", "override"])
@pytest.mark.parametrize("unsafe", [False, True], ids=["safe", "unsafe"])
def test_run(
    runner: CliRunner,
    unsafe: bool,
    loop: str | None,
    monkeypatch: MonkeyPatch,
    tmp_path: Path,
) -> None:
    if unsafe:
        component_class = "!!python/name:{0.__module__}.{0.__name__}".format(
            DummyComponent
        )
    else:
        component_class = "{0.__module__}:{0.__name__}".format(DummyComponent)

    monkeypatch.setenv("MYENVVAR", "from environment")
    tmp_path = tmp_path.joinpath("tmpfile")
    tmp_path.write_text("Hello, World!")

    config = """\
---
event_loop_policy: bogus
component:
  type: {cls}
  dummyval1: testval
  envval: !Env MYENVVAR
  textfileval: !TextFile {tmppath}
  binaryfileval: !BinaryFile {tmppath}
logging:
  version: 1
  disable_existing_loggers: false
""".format(
        cls=component_class, tmppath=str(tmp_path)
    )
    args = ["test.yml"]
    if unsafe:
        args.append("--unsafe")
    if loop:
        args.extend(["--loop", loop])

    with runner.isolated_filesystem(), patch(
        "asphalt.core.cli.run_application"
    ) as run_app:
        Path("test.yml").write_text(config)
        result = runner.invoke(cli.run, args)

        assert result.exit_code == 0
        assert run_app.call_count == 1
        args, kwargs = run_app.call_args
        assert len(args) == 0
        assert kwargs == {
            "component": {
                "type": DummyComponent if unsafe else component_class,
                "dummyval1": "testval",
                "envval": "from environment",
                "textfileval": "Hello, World!",
                "binaryfileval": b"Hello, World!",
            },
            "event_loop_policy": loop or "bogus",
            "logging": {"version": 1, "disable_existing_loggers": False},
        }


def test_run_multiple_configs(runner: CliRunner) -> None:
    component_class = "{0.__module__}:{0.__name__}".format(DummyComponent)
    config1 = """\
---
component:
  type: {}
  dummyval1: testval
logging:
  version: 1
  disable_existing_loggers: false
""".format(
        component_class
    )
    config2 = """\
---
component:
  dummyval1: alternate
  dummyval2: 10
"""

    with runner.isolated_filesystem(), patch(
        "asphalt.core.cli.run_application"
    ) as run_app:
        Path("conf1.yml").write_text(config1)
        Path("conf2.yml").write_text(config2)
        result = runner.invoke(cli.run, ["conf1.yml", "conf2.yml"])

        assert result.exit_code == 0
        assert run_app.call_count == 1
        args, kwargs = run_app.call_args
        assert len(args) == 0
        assert kwargs == {
            "component": {
                "type": component_class,
                "dummyval1": "alternate",
                "dummyval2": 10,
            },
            "logging": {"version": 1, "disable_existing_loggers": False},
        }


class TestServices:
    def write_config(self) -> None:
        Path("config.yml").write_text(
            """\
---
max_threads: 15
services:
  server:
    max_threads: 30
    component:
      type: myproject.server.ServerComponent
      components:
        wamp: &wamp
          host: wamp.example.org
          port: 8000
          tls: true
          auth_id: serveruser
          auth_secret: serverpass
        mailer:
          backend: smtp
  client:
    component:
      type: myproject.client.ClientComponent
      components:
        wamp:
          <<: *wamp
          auth_id: clientuser
          auth_secret: clientpass
logging:
  version: 1
  disable_existing_loggers: false
"""
        )

    @pytest.mark.parametrize("service", ["server", "client"])
    def test_run_service(self, runner: CliRunner, service: str) -> None:
        with runner.isolated_filesystem(), patch(
            "asphalt.core.cli.run_application"
        ) as run_app:
            self.write_config()
            result = runner.invoke(cli.run, ["-s", service, "config.yml"])

            assert result.exit_code == 0
            assert run_app.call_count == 1
            args, kwargs = run_app.call_args
            assert len(args) == 0
            if service == "server":
                assert kwargs == {
                    "max_threads": 30,
                    "component": {
                        "type": "myproject.server.ServerComponent",
                        "components": {
                            "wamp": {
                                "host": "wamp.example.org",
                                "port": 8000,
                                "tls": True,
                                "auth_id": "serveruser",
                                "auth_secret": "serverpass",
                            },
                            "mailer": {"backend": "smtp"},
                        },
                    },
                    "logging": {"version": 1, "disable_existing_loggers": False},
                }
            else:
                assert kwargs == {
                    "max_threads": 15,
                    "component": {
                        "type": "myproject.client.ClientComponent",
                        "components": {
                            "wamp": {
                                "host": "wamp.example.org",
                                "port": 8000,
                                "tls": True,
                                "auth_id": "clientuser",
                                "auth_secret": "clientpass",
                            }
                        },
                    },
                    "logging": {"version": 1, "disable_existing_loggers": False},
                }

    def test_service_not_found(self, runner: CliRunner) -> None:
        with runner.isolated_filesystem(), patch(
            "asphalt.core.cli.run_application"
        ) as run_app:
            self.write_config()
            result = runner.invoke(cli.run, ["-s", "foobar", "config.yml"])

            assert result.exit_code == 1
            assert run_app.call_count == 0
            assert result.output == "Error: Service 'foobar' has not been defined\n"

    def test_no_service_selected(self, runner: CliRunner) -> None:
        with runner.isolated_filesystem(), patch(
            "asphalt.core.cli.run_application"
        ) as run_app:
            self.write_config()
            result = runner.invoke(cli.run, ["config.yml"])

            assert result.exit_code == 1
            assert run_app.call_count == 0
            assert result.output == (
                "Error: Multiple services present in configuration file but no default service "
                "has been defined and no service was explicitly selected with -s / --service\n"
            )

    def test_bad_services_type(self, runner: CliRunner) -> None:
        with runner.isolated_filesystem(), patch(
            "asphalt.core.cli.run_application"
        ) as run_app:
            Path("config.yml").write_text(
                """\
---
services: blah
logging:
  version: 1
  disable_existing_loggers: false
"""
            )
            result = runner.invoke(cli.run, ["config.yml"])

            assert result.exit_code == 1
            assert run_app.call_count == 0
            assert (
                result.output == 'Error: The "services" key must be a dict, not str\n'
            )

    def test_no_services_defined(self, runner: CliRunner) -> None:
        with runner.isolated_filesystem(), patch(
            "asphalt.core.cli.run_application"
        ) as run_app:
            Path("config.yml").write_text(
                """\
---
services: {}
logging:
  version: 1
  disable_existing_loggers: false
"""
            )
            result = runner.invoke(cli.run, ["config.yml"])

            assert result.exit_code == 1
            assert run_app.call_count == 0
            assert result.output == "Error: No services have been defined\n"

    def test_run_only_service(self, runner) -> None:
        with runner.isolated_filesystem(), patch(
            "asphalt.core.cli.run_application"
        ) as run_app:
            Path("config.yml").write_text(
                """\
---
services:
  whatever:
    component:
      type: myproject.client.ClientComponent
logging:
  version: 1
  disable_existing_loggers: false
"""
            )
            result = runner.invoke(cli.run, ["config.yml"])

            assert result.exit_code == 0
            assert run_app.call_count == 1
            args, kwargs = run_app.call_args
            assert len(args) == 0
            assert kwargs == {
                "component": {"type": "myproject.client.ClientComponent"},
                "logging": {"version": 1, "disable_existing_loggers": False},
            }

    def test_run_default_service(self, runner: CliRunner) -> None:
        with runner.isolated_filesystem(), patch(
            "asphalt.core.cli.run_application"
        ) as run_app:
            Path("config.yml").write_text(
                """\
---
services:
  whatever:
    component:
      type: myproject.client.ClientComponent
  default:
    component:
      type: myproject.server.ServerComponent
logging:
  version: 1
  disable_existing_loggers: false
"""
            )
            result = runner.invoke(cli.run, ["config.yml"])

            assert result.exit_code == 0
            assert run_app.call_count == 1
            args, kwargs = run_app.call_args
            assert len(args) == 0
            assert kwargs == {
                "component": {"type": "myproject.server.ServerComponent"},
                "logging": {"version": 1, "disable_existing_loggers": False},
            }

    def test_service_env_variable(
        self, runner: CliRunner, monkeypatch: MonkeyPatch
    ) -> None:
        with runner.isolated_filesystem(), patch(
            "asphalt.core.cli.run_application"
        ) as run_app:
            Path("config.yml").write_text(
                """\
---
services:
  whatever:
    component:
      type: myproject.client.ClientComponent
  default:
    component:
      type: myproject.server.ServerComponent
logging:
  version: 1
  disable_existing_loggers: false
"""
            )
            monkeypatch.setenv("ASPHALT_SERVICE", "whatever")
            result = runner.invoke(cli.run, ["config.yml"])

            assert result.exit_code == 0
            assert run_app.call_count == 1
            args, kwargs = run_app.call_args
            assert len(args) == 0
            assert kwargs == {
                "component": {"type": "myproject.client.ClientComponent"},
                "logging": {"version": 1, "disable_existing_loggers": False},
            }

    def test_service_env_variable_override(
        self, runner: CliRunner, monkeypatch: MonkeyPatch
    ) -> None:
        with runner.isolated_filesystem(), patch(
            "asphalt.core.cli.run_application"
        ) as run_app:
            Path("config.yml").write_text(
                """\
---
services:
  whatever:
    component:
      type: myproject.client.ClientComponent
  default:
    component:
      type: myproject.server.ServerComponent
logging:
  version: 1
  disable_existing_loggers: false
"""
            )
            monkeypatch.setenv("ASPHALT_SERVICE", "whatever")
            result = runner.invoke(cli.run, ["-s", "default", "config.yml"])

            assert result.exit_code == 0
            assert run_app.call_count == 1
            args, kwargs = run_app.call_args
            assert len(args) == 0
            assert kwargs == {
                "component": {"type": "myproject.server.ServerComponent"},
                "logging": {"version": 1, "disable_existing_loggers": False},
            }



"""This is the change detector component for the Asphalt webnotifier tutorial."""
# isort: off
from __future__ import annotations

import asyncio
import logging
from typing import Any, AsyncIterator

import aiohttp
from asphalt.core import Component, Event, Signal, context_teardown, Context

logger = logging.getLogger(__name__)


class WebPageChangeEvent(Event):
    def __init__(self, source, topic, old_lines, new_lines):
        super().__init__(source, topic)
        self.old_lines = old_lines
        self.new_lines = new_lines


class Detector:
    changed = Signal(WebPageChangeEvent)

    def __init__(self, url: str, delay: float):
        self.url = url
        self.delay = delay

    async def run(self) -> None:
        async with aiohttp.ClientSession() as session:
            last_modified, old_lines = None, None
            while True:
                logger.debug("Fetching contents of %s", self.url)
                headers: dict[str, Any] = (
                    {"if-modified-since": last_modified} if last_modified else {}
                )
                async with session.get(self.url, headers=headers) as resp:
                    logger.debug("Response status: %d", resp.status)
                    if resp.status == 200:
                        last_modified = resp.headers["date"]
                        new_lines = (await resp.text()).split("\n")
                        if old_lines is not None and old_lines != new_lines:
                            self.changed.dispatch(old_lines, new_lines)

                        old_lines = new_lines

                await asyncio.sleep(self.delay)


class ChangeDetectorComponent(Component):
    def __init__(self, url: str, delay: int = 10):
        self.url = url
        self.delay = delay

    @context_teardown
    async def start(self, ctx: Context) -> AsyncIterator[None]:
        detector = Detector(self.url, self.delay)
        ctx.add_resource(detector)
        task = asyncio.create_task(detector.run())
        logging.info(
            'Started web page change detector for url "%s" with a delay of %d seconds',
            self.url,
            self.delay,
        )

        yield

        # This part is run when the context is finished
        task.cancel()
        await asyncio.gather(task, return_exceptions=True)
        logging.info("Shut down web page change detector")



from __future__ import annotations

import gc
from asyncio import AbstractEventLoop, Queue, all_tasks, current_task
from datetime import datetime, timedelta, timezone
from typing import NoReturn

import pytest
from async_generator import aclosing

from asphalt.core import Event, Signal, stream_events, wait_event


class DummyEvent(Event):
    def __init__(self, source: DummySource, topic: str, *args, **kwargs):
        super().__init__(source, topic)
        self.args = args
        self.kwargs = kwargs


class DummySource:
    event_a = Signal(DummyEvent)
    event_b = Signal(DummyEvent)


@pytest.fixture
def source() -> DummySource:
    return DummySource()


class TestEvent:
    def test_utc_timestamp(self, source: DummySource) -> None:
        timestamp = datetime.now(timezone(timedelta(hours=2)))
        event = Event(source, "sometopic", timestamp.timestamp())
        assert event.utc_timestamp == timestamp
        assert event.utc_timestamp.tzinfo == timezone.utc

    def test_event_repr(self, source: DummySource) -> None:
        event = Event(source, "sometopic")
        assert repr(event) == f"Event(source={source!r}, topic='sometopic')"


class TestSignal:
    def test_class_attribute_access(self) -> None:
        """
        Test that accessing the descriptor on the class level returns the same signal instance.

        """
        signal = Signal(DummyEvent)

        class EventSource:
            dummysignal = signal

        assert EventSource.dummysignal is signal

    @pytest.mark.asyncio
    async def test_disconnect(self, source: DummySource) -> None:
        """Test that an event listener no longer receives events after it's been removed."""
        events: list[DummyEvent] = []
        source.event_a.connect(events.append)
        assert await source.event_a.dispatch(1)

        source.event_a.disconnect(events.append)
        assert await source.event_a.dispatch(2)

        assert len(events) == 1
        assert events[0].args == (1,)

    def test_disconnect_nonexistent_listener(self, source: DummySource) -> None:
        """Test that attempting to remove a nonexistent event listener will not raise an error."""
        source.event_a.connect(lambda event: None)
        source.event_a.disconnect(lambda event: None)

    def test_disconnect_no_listeners(self, source: DummySource) -> None:
        """
        Test that disconnecting a nonexistent listener while listeners is None will still work.

        """
        source.event_a.disconnect(lambda event: None)

    @pytest.mark.asyncio
    async def test_dispatch_event_coroutine(self, source: DummySource) -> None:
        """Test that a coroutine function can be an event listener."""

        async def callback(event: Event) -> None:
            events.append(event)

        events: list[DummyEvent] = []
        source.event_a.connect(callback)
        assert await source.event_a.dispatch("x", "y", a=1, b=2)

        assert len(events) == 1
        assert events[0].args == ("x", "y")
        assert events[0].kwargs == {"a": 1, "b": 2}

    @pytest.mark.asyncio
    async def test_dispatch_raw(self, source: DummySource) -> None:
        """Test that dispatch_raw() correctly dispatches the given event."""
        events: list[DummyEvent] = []
        source.event_a.connect(events.append)
        event = DummyEvent(source, "event_a", "x", "y", a=1, b=2)
        assert await source.event_a.dispatch_raw(event)

        assert events == [event]

    @pytest.mark.asyncio
    async def test_dispatch_log_exceptions(
        self, event_loop: AbstractEventLoop, source: DummySource, caplog
    ) -> None:
        """Test that listener exceptions are logged and that dispatch() resolves to ``False``."""

        def listener(event) -> NoReturn:
            raise Exception("regular")

        async def coro_listener(event) -> NoReturn:
            raise Exception("coroutine")

        source.event_a.connect(listener)
        source.event_a.connect(coro_listener)
        assert not await source.event_a.dispatch()

        assert len(caplog.records) == 2
        for record in caplog.records:
            assert "Uncaught exception in event listener" in record.message

    @pytest.mark.asyncio
    async def test_dispatch_event_no_listeners(self, source: DummySource) -> None:
        """Test that dispatching an event when there are no listeners will still work."""
        assert await source.event_a.dispatch()

    @pytest.mark.asyncio
    async def test_dispatch_event_cancel(self, source: DummySource) -> None:
        """Test that dispatching an event when there are no listeners will still work."""
        source.event_a.connect(lambda event: None)
        future = source.event_a.dispatch()
        future.cancel()
        task = next(t for t in all_tasks() if t is not current_task())
        await task

    @pytest.mark.asyncio
    async def test_connect_twice(self, source: DummySource) -> None:
        """Test that if the same callback is connected twice, the second connect is a no-op."""
        events: list[DummyEvent] = []
        source.event_a.connect(events.append)
        source.event_a.connect(events.append)
        assert await source.event_a.dispatch()

        assert len(events) == 1

    @pytest.mark.asyncio
    async def test_dispatch_raw_class_mismatch(self, source: DummySource) -> None:
        """Test that passing an event of the wrong type raises an AssertionError."""
        with pytest.raises(TypeError) as exc:
            await source.event_a.dispatch_raw(Event(source, "event_a"))

        assert str(exc.value) == "event must be of type test_event.DummyEvent"

    @pytest.mark.asyncio
    async def test_wait_event(self, source: DummySource, event_loop) -> None:
        event_loop.call_soon(source.event_a.dispatch)
        received_event = await source.event_a.wait_event()
        assert received_event.topic == "event_a"

    @pytest.mark.parametrize(
        "filter, expected_values",
        [(None, [1, 2, 3]), (lambda event: event.args[0] in (3, None), [3])],
        ids=["nofilter", "filter"],
    )
    @pytest.mark.asyncio
    async def test_stream_events(
        self, source: DummySource, filter, expected_values
    ) -> None:
        values = []
        async with aclosing(source.event_a.stream_events(filter)) as stream:
            for i in range(1, 4):
                source.event_a.dispatch(i)

            source.event_a.dispatch(None)

            async for event in stream:
                if event.args[0] is not None:
                    values.append(event.args[0])
                else:
                    break

        assert values == expected_values

    def test_memory_leak(self) -> None:
        """
        Test that activating a Signal does not prevent its owner object from being garbage
        collected.

        """

        class SignalOwner:
            dummy = Signal(Event)

        owner = SignalOwner()
        owner.dummy
        del owner
        gc.collect()  # needed on PyPy
        assert (
            next((x for x in gc.get_objects() if isinstance(x, SignalOwner)), None)
            is None
        )


@pytest.mark.parametrize(
    "filter, expected_value",
    [(None, 1), (lambda event: event.args[0] == 3, 3)],
    ids=["nofilter", "filter"],
)
@pytest.mark.asyncio
async def test_wait_event(event_loop, filter, expected_value) -> None:
    """
    Test that wait_event returns the first event matched by the filter, or the first event if there
    is no filter.

    """
    source1 = DummySource()
    for i in range(1, 4):
        event_loop.call_soon(source1.event_a.dispatch, i)

    event = await wait_event([source1.event_a], filter)
    assert event.args == (expected_value,)


@pytest.mark.parametrize(
    "filter, expected_values",
    [(None, [1, 2, 3, 1, 2, 3]), (lambda event: event.args[0] in (3, None), [3, 3])],
    ids=["nofilter", "filter"],
)
@pytest.mark.asyncio
async def test_stream_events(filter, expected_values) -> None:
    source1, source = DummySource(), DummySource()
    values = []
    async with aclosing(
        stream_events([source1.event_a, source.event_b], filter)
    ) as stream:
        for signal in [source1.event_a, source.event_b]:
            for i in range(1, 4):
                signal.dispatch(i)

        source1.event_a.dispatch(None)

        async for event in stream:
            if event.args[0] is not None:
                values.append(event.args[0])
            else:
                break

    assert values == expected_values


@pytest.mark.asyncio
async def test_stream_events_memleak() -> None:
    """Test that closing but never iterating the event stream will not cause a memory leak."""
    source = DummySource()
    gc.collect()
    gc.collect()
    num_queues_before = len([x for x in gc.get_objects() if type(x) is Queue])
    async with aclosing(stream_events([source.event_a])):
        pass

    gc.collect()
    gc.collect()
    num_queues_after = len([x for x in gc.get_objects() if type(x) is Queue])
    assert num_queues_after == num_queues_before



from __future__ import annotations

import asyncio
import logging
import sys
from asyncio import AbstractEventLoop
from concurrent.futures import ThreadPoolExecutor
from typing import NoReturn
from unittest.mock import patch

import pytest
from _pytest.logging import LogCaptureFixture
from _pytest.monkeypatch import MonkeyPatch

from asphalt.core.component import CLIApplicationComponent, Component
from asphalt.core.context import Context
from asphalt.core.runner import run_application, sigterm_handler


class ShutdownComponent(Component):
    def __init__(self, method: str = "stop"):
        self.method = method
        self.teardown_callback_called = False
        self.exception = None

    def teardown_callback(self, exception) -> None:
        self.teardown_callback_called = True
        self.exception = exception

    def press_ctrl_c(self) -> NoReturn:
        raise KeyboardInterrupt

    async def start(self, ctx: Context) -> None:
        ctx.add_teardown_callback(self.teardown_callback, pass_exception=True)

        if self.method == "stop":
            ctx.loop.call_later(0.1, ctx.loop.stop)
        elif self.method == "exit":
            ctx.loop.call_later(0.1, sys.exit)
        elif self.method == "keyboard":
            ctx.loop.call_later(0.1, self.press_ctrl_c)
        elif self.method == "sigterm":
            ctx.loop.call_later(
                0.1, sigterm_handler, logging.getLogger(__name__), ctx.loop
            )
        elif self.method == "exception":
            raise RuntimeError("this should crash the application")
        elif self.method == "timeout":
            await asyncio.sleep(1)


class DummyCLIApp(CLIApplicationComponent):
    async def run(self, ctx: Context) -> int:
        return 20


@pytest.fixture(autouse=True)
def prevent_logging_shutdown(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setattr("asphalt.core.runner.shutdown", lambda: None)


def test_sigterm_handler_loop_not_running(event_loop: AbstractEventLoop) -> None:
    """Test that the SIGTERM handler does nothing if the event loop is not running."""
    sigterm_handler(logging.getLogger(__name__), event_loop)


@pytest.mark.parametrize(
    "logging_config",
    [None, logging.INFO, {"version": 1, "loggers": {"asphalt": {"level": "INFO"}}}],
    ids=["disabled", "loglevel", "dictconfig"],
)
def test_run_logging_config(event_loop: AbstractEventLoop, logging_config) -> None:
    """Test that logging initialization happens as expected."""
    with patch("asphalt.core.runner.basicConfig") as basicConfig, patch(
        "asphalt.core.runner.dictConfig"
    ) as dictConfig:
        run_application(ShutdownComponent(), logging=logging_config)

    assert basicConfig.call_count == (1 if logging_config == logging.INFO else 0)
    assert dictConfig.call_count == (1 if isinstance(logging_config, dict) else 0)


@pytest.mark.parametrize("max_threads", [None, 3])
def test_run_max_threads(
    event_loop: AbstractEventLoop, max_threads: int | None
) -> None:
    """
    Test that a new default executor is installed if and only if the max_threads argument is given.

    """
    component = ShutdownComponent()
    with patch("asphalt.core.runner.ThreadPoolExecutor") as mock_executor:
        mock_executor.configure_mock(
            side_effect=lambda *args, **kwargs: ThreadPoolExecutor(*args, **kwargs)
        )
        run_application(component, max_threads=max_threads)

    if max_threads:
        mock_executor.assert_called_once_with(max_threads)
    else:
        assert not mock_executor.called


def test_uvloop_policy(caplog: LogCaptureFixture) -> None:
    """Test that the runner switches to a different event loop policy when instructed to."""
    pytest.importorskip("uvloop", reason="uvloop not installed")
    caplog.set_level(logging.INFO)
    component = ShutdownComponent()
    old_policy = asyncio.get_event_loop_policy()
    run_application(component, event_loop_policy="uvloop")
    asyncio.set_event_loop_policy(old_policy)

    records = [
        record for record in caplog.records if record.name == "asphalt.core.runner"
    ]
    assert len(records) == 6
    assert records[0].message == "Running in development mode"
    assert records[1].message == "Switched event loop policy to uvloop.EventLoopPolicy"
    assert records[2].message == "Starting application"
    assert records[3].message == "Application started"
    assert records[4].message == "Stopping application"
    assert records[5].message == "Application stopped"


def test_run_callbacks(
    event_loop: AbstractEventLoop, caplog: LogCaptureFixture
) -> None:
    """
    Test that the teardown callbacks are run when the application is started and shut down properly
    and that the proper logging messages are emitted.

    """
    caplog.set_level(logging.INFO)
    component = ShutdownComponent()
    run_application(component)

    assert component.teardown_callback_called
    records = [
        record for record in caplog.records if record.name == "asphalt.core.runner"
    ]
    assert len(records) == 5
    assert records[0].message == "Running in development mode"
    assert records[1].message == "Starting application"
    assert records[2].message == "Application started"
    assert records[3].message == "Stopping application"
    assert records[4].message == "Application stopped"


@pytest.mark.parametrize("method", ["exit", "keyboard", "sigterm"])
def test_clean_exit(
    event_loop: AbstractEventLoop, caplog: LogCaptureFixture, method: str
) -> None:
    """
    Test that when Ctrl+C is pressed during event_loop.run_forever(), run_application() exits
    cleanly.

    """
    caplog.set_level(logging.INFO)
    component = ShutdownComponent(method=method)
    run_application(component)

    records = [
        record for record in caplog.records if record.name == "asphalt.core.runner"
    ]
    assert len(records) == 5
    assert records[0].message == "Running in development mode"
    assert records[1].message == "Starting application"
    assert records[2].message == "Application started"
    assert records[3].message == "Stopping application"
    assert records[4].message == "Application stopped"


def test_run_start_exception(
    event_loop: AbstractEventLoop, caplog: LogCaptureFixture
) -> None:
    """
    Test that an exception caught during the application initialization is put into the
    application context and made available to teardown callbacks.

    """
    caplog.set_level(logging.INFO)
    component = ShutdownComponent(method="exception")
    pytest.raises(SystemExit, run_application, component)

    assert str(component.exception) == "this should crash the application"
    records = [
        record for record in caplog.records if record.name == "asphalt.core.runner"
    ]
    assert len(records) == 5
    assert records[0].message == "Running in development mode"
    assert records[1].message == "Starting application"
    assert records[2].message == "Error during application startup"
    assert records[3].message == "Stopping application"
    assert records[4].message == "Application stopped"


def test_run_start_timeout(
    event_loop: AbstractEventLoop, caplog: LogCaptureFixture
) -> None:
    """
    Test that when the root component takes too long to start up, the runner exits and logs the
    appropriate error message.

    """
    caplog.set_level(logging.INFO)
    component = ShutdownComponent(method="timeout")
    pytest.raises(SystemExit, run_application, component, start_timeout=1)

    records = [
        record for record in caplog.records if record.name == "asphalt.core.runner"
    ]
    assert len(records) == 5
    assert records[0].message == "Running in development mode"
    assert records[1].message == "Starting application"
    assert records[2].message == "Timeout waiting for the root component to start"
    assert records[3].message == "Stopping application"
    assert records[4].message == "Application stopped"


def test_dict_config(event_loop: AbstractEventLoop, caplog: LogCaptureFixture) -> None:
    """Test that component configuration passed as a dictionary works."""
    caplog.set_level(logging.INFO)
    component_class = "{0.__module__}:{0.__name__}".format(ShutdownComponent)
    run_application(component={"type": component_class})

    records = [
        record for record in caplog.records if record.name == "asphalt.core.runner"
    ]
    assert len(records) == 5
    assert records[0].message == "Running in development mode"
    assert records[1].message == "Starting application"
    assert records[2].message == "Application started"
    assert records[3].message == "Stopping application"
    assert records[4].message == "Application stopped"


def test_run_cli_application(
    event_loop: AbstractEventLoop, caplog: LogCaptureFixture
) -> None:
    caplog.set_level(logging.INFO)
    with pytest.raises(SystemExit) as exc:
        run_application(DummyCLIApp())

    assert exc.value.code == 20

    records = [
        record for record in caplog.records if record.name == "asphalt.core.runner"
    ]
    assert len(records) == 5
    assert records[0].message == "Running in development mode"
    assert records[1].message == "Starting application"
    assert records[2].message == "Application started"
    assert records[3].message == "Stopping application"
    assert records[4].message == "Application stopped"



from __future__ import annotations

import os
from pathlib import Path
from typing import Any, Dict, Optional

import click
from ruamel.yaml import YAML, ScalarNode
from ruamel.yaml.loader import Loader

from asphalt.core.runner import policies, run_application
from asphalt.core.utils import merge_config, qualified_name


def env_constructor(loader: Loader, node: ScalarNode) -> str | None:
    value = loader.construct_scalar(node)
    return os.getenv(value)


def text_file_constructor(loader: Loader, node: ScalarNode) -> str:
    value = loader.construct_scalar(node)
    return Path(value).read_text()


def binary_file_constructor(loader: Loader, node: ScalarNode) -> bytes:
    value = loader.construct_scalar(node)
    return Path(value).read_bytes()


@click.group()
def main() -> None:
    pass  # pragma: no cover


@main.command(help="Read one or more configuration files and start the application.")
@click.argument("configfile", type=click.File(), nargs=-1, required=True)
@click.option(
    "--unsafe",
    is_flag=True,
    default=False,
    help="use unsafe mode when loading YAML (enables markup extensions)",
)
@click.option(
    "-l",
    "--loop",
    type=click.Choice(policies.names),
    help="alternate event loop policy",
)
@click.option(
    "-s",
    "--service",
    type=str,
    help="service to run (if the configuration file contains multiple services)",
)
def run(configfile, unsafe: bool, loop: Optional[str], service: Optional[str]) -> None:
    yaml = YAML(typ="unsafe" if unsafe else "safe")
    yaml.constructor.add_constructor("!Env", env_constructor)
    yaml.constructor.add_constructor("!TextFile", text_file_constructor)
    yaml.constructor.add_constructor("!BinaryFile", binary_file_constructor)

    # Read the configuration from the supplied YAML files
    config: Dict[str, Any] = {}
    for path in configfile:
        config_data = yaml.load(path)
        assert isinstance(
            config_data, dict
        ), "the document root element must be a dictionary"
        config = merge_config(config, config_data)

    # Override the event loop policy if specified
    if loop:
        config["event_loop_policy"] = loop

    services = config.pop("services", {})
    if not isinstance(services, dict):
        raise click.ClickException(
            f'The "services" key must be a dict, not {qualified_name(services)}'
        )

    # If "component" was defined, use that as the default service if one has not been defined yet
    if "component" in config:
        component = config.pop("component")
        services.setdefault("default", dict(component=component))

    # Try to figure out which service to launch
    service = service or os.getenv("ASPHALT_SERVICE")
    if len(services) == 0:
        raise click.ClickException("No services have been defined")
    elif service:
        try:
            service_config = services[service]
        except KeyError:
            raise click.ClickException(
                f"Service {service!r} has not been defined"
            ) from None
    elif len(services) == 1:
        service_config = next(iter(services.values()))
    elif "default" in services:
        service_config = services["default"]
    else:
        raise click.ClickException(
            "Multiple services present in configuration file but no default service has been "
            "defined and no service was explicitly selected with -s / --service"
        )

    # Merge the service-level configuration with the top level one
    config = merge_config(config, service_config)

    # Start the application
    run_application(**config)



from __future__ import annotations

__all__ = ("Event", "Signal", "wait_event", "stream_events")

import logging
import sys
import weakref
from asyncio import Queue, create_task, get_running_loop, iscoroutine, wait
from datetime import datetime, timezone
from inspect import getmembers, isawaitable
from time import time as stdlib_time
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Generic,
    List,
    MutableMapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    cast,
)
from weakref import WeakKeyDictionary

from typeguard import check_argument_types

from asphalt.core.utils import qualified_name

if sys.version_info >= (3, 10):
    from contextlib import aclosing
else:
    from async_generator import aclosing

logger = logging.getLogger(__name__)


class Event:
    """
    The base class for all events.

    :param source: the object where this event originated from
    :param topic: the event topic
    :param time: the time the event occurred

    :ivar source: the object where this event originated from
    :ivar str topic: the topic
    :ivar float time: event creation time as seconds from the UNIX epoch
    """

    __slots__ = "source", "topic", "time"

    def __init__(self, source: Any, topic: str, time: float | None = None) -> None:
        self.source = source
        self.topic = topic
        self.time = time or stdlib_time()

    @property
    def utc_timestamp(self) -> datetime:
        """
        Return a timezone aware :class:`~datetime.datetime` corresponding to the ``time`` variable,
        using the UTC timezone.

        """
        return datetime.fromtimestamp(self.time, timezone.utc)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(source={self.source!r}, topic={self.topic!r})"
        )


T_Event = TypeVar("T_Event", bound=Event)


class Signal(Generic[T_Event]):
    """
    Declaration of a signal that can be used to dispatch events.

    This is a descriptor that returns itself on class level attribute access and a bound version of
    itself on instance level access. Connecting listeners and dispatching events only works with
    these bound instances.

    Each signal must be assigned to a class attribute, but only once. The Signal will not function
    correctly if the same Signal instance is assigned to multiple attributes.

    :param event_class: an event class
    """

    __slots__ = "event_class", "topic", "source", "listeners", "bound_signals"

    def __init__(
        self,
        event_class: Type[T_Event],
        *,
        source: Any = None,
        topic: Optional[str] = None,
    ) -> None:
        assert check_argument_types()
        self.event_class = event_class
        self.topic = topic
        if source is not None:
            self.source = weakref.ref(source)
            self.listeners: Optional[List[Callable]] = None
        else:
            assert issubclass(
                event_class, Event
            ), "event_class must be a subclass of Event"
            self.bound_signals: MutableMapping[Any, Signal] = WeakKeyDictionary()

    def __get__(self, instance, owner) -> Signal:
        if instance is None:
            return self

        # Find the attribute this Signal was assigned to (needed only on Python 3.5)
        if self.topic is None:
            self.topic = next(
                attr for attr, value in getmembers(owner, lambda value: value is self)
            )

        try:
            return self.bound_signals[instance]
        except KeyError:
            bound_signal = Signal(self.event_class, source=instance, topic=self.topic)
            self.bound_signals[instance] = bound_signal
            return bound_signal

    def __set_name__(self, owner, name: str) -> None:
        self.topic = name

    def connect(self, callback: Callable[[T_Event], Any]) -> Callable[[T_Event], Any]:
        """
        Connect a callback to this signal.

        Each callable can only be connected once. Duplicate registrations are ignored.

        If you need to pass extra arguments to the callback, you can use :func:`functools.partial`
        to wrap the callable.

        :param callback: a callable that will receive an event object as its only argument.
        :return: the value of ``callback`` argument

        """
        assert check_argument_types()
        if self.listeners is None:
            self.listeners = []
        if callback not in self.listeners:
            self.listeners.append(callback)

        return callback

    def disconnect(self, callback: Callable) -> None:
        """
        Disconnects the given callback.

        The callback will no longer receive events from this signal.

        No action is taken if the callback is not on the list of listener callbacks.

        :param callback: the callable to remove

        """
        assert check_argument_types()
        try:
            if self.listeners is not None:
                self.listeners.remove(callback)
        except ValueError:
            pass

    def dispatch_raw(self, event: Event) -> Awaitable[bool]:
        """
        Dispatch the given event object to all listeners.

        Creates a new task in which all listener callbacks are called with the given event as
        the only argument. Coroutine callbacks are converted to their own respective tasks and
        waited for concurrently.

        Before the dispatching is done, a snapshot of the listeners is taken and the event is only
        dispatched to those listeners, so adding a listener between the call to this method and the
        actual dispatching will only affect future calls to this method.

        :param event: the event object to dispatch
        :returns: an awaitable that completes when all the callbacks have been called (and any
            awaitables waited on) and resolves to ``True`` if there were no exceptions raised by
            the callbacks, ``False`` otherwise

        """

        async def do_dispatch() -> None:
            awaitables: List[Awaitable[Any]] = []
            all_successful = True
            for callback in listeners:
                try:
                    retval = callback(event)
                except Exception:
                    logger.exception("Uncaught exception in event listener")
                    all_successful = False
                else:
                    if iscoroutine(retval):
                        awaitables.append(create_task(retval))
                    elif isawaitable(retval):
                        awaitables.append(retval)

            # For any callbacks that returned awaitables, wait for their completion and log any
            # exceptions they raised
            if awaitables:
                done, _ = await wait(awaitables)
                for f in done:
                    exc = f.exception()
                    if exc is not None:
                        all_successful = False
                        logger.error(
                            "Uncaught exception in event listener", exc_info=exc
                        )

            if not future.cancelled():
                future.set_result(all_successful)

        if not isinstance(event, self.event_class):
            raise TypeError(f"event must be of type {qualified_name(self.event_class)}")

        loop = get_running_loop()
        future = loop.create_future()
        if self.listeners:
            listeners = list(self.listeners)
            loop.create_task(do_dispatch())
        else:
            future.set_result(True)

        return future

    def dispatch(self, *args, **kwargs) -> Awaitable[bool]:
        """
        Create and dispatch an event.

        This method constructs an event object and then passes it to :meth:`dispatch_event` for
        the actual dispatching.

        :param args: positional arguments to the constructor of the associated event class
        :param kwargs: keyword arguments to the constructor of the associated event class
        :returns: an awaitable that completes when all the callbacks have been called (and any
            awaitables waited on) and resolves to ``True`` if there were no exceptions raised by
            the callbacks, ``False`` otherwise

        """
        event = self.event_class(self.source(), cast(str, self.topic), *args, **kwargs)
        return self.dispatch_raw(event)

    def wait_event(
        self, filter: Callable[[T_Event], bool] = None
    ) -> Awaitable[T_Event]:
        """Shortcut for calling :func:`wait_event` with this signal in the first argument."""
        return wait_event([self], filter)

    def stream_events(
        self, filter: Callable[[T_Event], bool] = None, *, max_queue_size: int = 0
    ) -> AsyncIterator[T_Event]:
        """Shortcut for calling :func:`stream_events` with this signal in the first argument."""
        return stream_events([self], filter, max_queue_size=max_queue_size)


def stream_events(
    signals: Sequence[Signal[T_Event]],
    filter: Callable[[T_Event], bool] = None,
    *,
    max_queue_size: int = 0,
) -> AsyncIterator[T_Event]:
    """
    Return an async generator that yields events from the given signals.

    Only events that pass the filter callable (if one has been given) are returned.
    If no filter function was given, all events are yielded from the generator.

    :param signals: the signals to get events from
    :param filter: a callable that takes an event object as an argument and returns ``True`` if
        the event should pass, ``False`` if not
    :param max_queue_size: maximum size of the queue, after which it will start to drop events

    """
    queue: Queue[T_Event] | None

    async def streamer() -> AsyncIterator[T_Event]:
        try:
            while queue is not None:
                event = await queue.get()
                if filter is None or filter(event):
                    yield event
        finally:
            cleanup()

    def cleanup() -> None:
        nonlocal queue
        if queue is not None:
            for signal in signals:
                signal.disconnect(queue.put_nowait)

            queue = None

    assert check_argument_types()
    queue = Queue(max_queue_size)
    for signal in signals:
        signal.connect(queue.put_nowait)

    gen = [streamer()]  # this is to allow the reference count to drop to 0
    weakref.finalize(gen[0], cleanup)
    return gen.pop()


async def wait_event(
    signals: Sequence[Signal[T_Event]], filter: Callable[[T_Event], bool] = None
) -> T_Event:
    """
    Wait until any of the given signals dispatches an event that satisfies the filter (if any).

    If no filter has been given, the first event dispatched from the signal is returned.

    :param signals: the signals to get events from
    :param filter: a callable that takes an event object as an argument and returns ``True`` if
        the event should pass, ``False`` if not
    :return: the event that was dispatched

    """
    if sys.version_info >= (3, 5, 3):
        assert check_argument_types()

    async with aclosing(stream_events(signals, filter)) as events:
        return await events.asend(None)



from __future__ import annotations

__all__ = (
    "ResourceEvent",
    "ResourceConflict",
    "ResourceNotFound",
    "NoCurrentContext",
    "TeardownError",
    "Context",
    "executor",
    "context_teardown",
    "current_context",
    "get_resource",
    "get_resources",
    "require_resource",
    "_Dependency",
    "inject",
    "resource",
)

import logging
import re
import sys
import types
import warnings
from asyncio import (
    AbstractEventLoop,
    current_task,
    get_event_loop,
    get_running_loop,
    iscoroutinefunction,
)
from collections.abc import Coroutine
from collections.abc import Sequence as ABCSequence
from concurrent.futures import Executor
from contextvars import ContextVar, Token, copy_context
from dataclasses import dataclass, field
from enum import Enum, auto
from functools import partial, wraps
from inspect import (
    Parameter,
    getattr_static,
    isasyncgenfunction,
    isawaitable,
    isclass,
    signature,
)
from traceback import format_exception
from types import TracebackType
from typing import (
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    get_type_hints,
    overload,
)

import asyncio_extras
from async_generator import async_generator
from typeguard import check_argument_types

from asphalt.core.event import Event, Signal, wait_event
from asphalt.core.utils import callable_name, qualified_name

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec

if sys.version_info >= (3, 8):
    from typing import get_args, get_origin
else:
    from typing_extensions import get_args, get_origin

logger = logging.getLogger(__name__)
factory_callback_type = Callable[["Context"], Any]
resource_name_re = re.compile(r"\w+")
T_Resource = TypeVar("T_Resource")
T_Retval = TypeVar("T_Retval")
T_Context = TypeVar("T_Context", bound="Context")
T_Self = TypeVar("T_Self")
P = ParamSpec("P")
_current_context: ContextVar[Context | None] = ContextVar(
    "_current_context", default=None
)


class ResourceContainer:
    """
    Contains the resource value or its factory callable, plus some metadata.

    :ivar value_or_factory: the resource value or the factory callback
    :ivar types: type names the resource was registered with
    :vartype types: Tuple[type, ...]
    :ivar str name: name of the resource
    :ivar str context_attr: the context attribute of the resource
    :ivar bool is_factory: ``True`` if ``value_or_factory`` if this is a resource factory
    """

    __slots__ = "value_or_factory", "types", "name", "context_attr", "is_factory"

    def __init__(
        self,
        value_or_factory: Any,
        types: Tuple[type, ...],
        name: str,
        context_attr: Optional[str],
        is_factory: bool,
    ) -> None:
        self.value_or_factory = value_or_factory
        self.types = types
        self.name = name
        self.context_attr = context_attr
        self.is_factory = is_factory

    def generate_value(self, ctx: Context) -> Any:
        assert self.is_factory, "generate_value() only works for resource factories"
        value = self.value_or_factory(ctx)

        container = ResourceContainer(
            value, self.types, self.name, self.context_attr, False
        )
        for type_ in self.types:
            ctx._resources[(type_, self.name)] = container

        if self.context_attr:
            setattr(ctx, self.context_attr, value)

        return value

    def __repr__(self) -> str:
        typenames = ", ".join(qualified_name(cls) for cls in self.types)
        value_repr = (
            "factory=%s" % callable_name(self.value_or_factory)
            if self.is_factory
            else "value=%r" % self.value_or_factory
        )
        return (
            "{self.__class__.__name__}({value_repr}, types=[{typenames}], name={self.name!r}, "
            "context_attr={self.context_attr!r})".format(
                self=self, value_repr=value_repr, typenames=typenames
            )
        )


class ResourceEvent(Event):
    """
    Dispatched when a resource or resource factory has been added to a context.

    :ivar resource_types: types the resource was registered under
    :vartype resource_types: Tuple[type, ...]
    :ivar str name: name of the resource
    :ivar bool is_factory: ``True`` if a resource factory was added, ``False`` if a regular
        resource was added
    """

    __slots__ = "resource_types", "resource_name", "is_factory"

    def __init__(
        self,
        source: Context,
        topic: str,
        types: Tuple[type, ...],
        name: str,
        is_factory: bool,
    ) -> None:
        super().__init__(source, topic)
        self.resource_types = types
        self.resource_name = name
        self.is_factory = is_factory


class ResourceConflict(Exception):
    """
    Raised when a new resource that is being published conflicts with an existing resource or
    context variable.
    """


class ResourceNotFound(LookupError):
    """Raised when a resource request cannot be fulfilled within the allotted time."""

    def __init__(self, type: type, name: str) -> None:
        super().__init__(type, name)
        self.type = type
        self.name = name

    def __str__(self):
        return "no matching resource was found for type={typename} name={self.name!r}".format(
            self=self, typename=qualified_name(self.type)
        )


class TeardownError(Exception):
    """
    Raised after context teardown when one or more teardown callbacks raised an exception.

    :ivar exceptions: exceptions raised during context teardown, in the order in which they were
        raised
    :vartype exceptions: List[Exception]
    """

    def __init__(self, exceptions: List[Exception]) -> None:
        super().__init__(exceptions)
        self.exceptions = exceptions

    def __str__(self):
        separator = "----------------------------\n"
        tracebacks = separator.join(
            "\n".join(format_exception(type(exc), exc, exc.__traceback__))
            for exc in self.exceptions
        )
        return "{} exceptions(s) were raised by teardown callbacks:\n{}{}".format(
            len(self.exceptions), separator, tracebacks
        )


class NoCurrentContext(Exception):
    """Raised by :func: `current_context` when there is no active context."""

    def __init__(self) -> None:
        super().__init__("There is no active context")


class ContextState(Enum):
    open = auto()
    closing = auto()
    closed = auto()


class Context:
    """
    Contexts give request handlers and callbacks access to resources.

    Contexts are stacked in a way that accessing an attribute that is not present in the current
    context causes the attribute to be looked up in the parent instance and so on, until the
    attribute is found (or :class:`AttributeError` is raised).

    :param parent: the parent context, if any

    :ivar Context parent: the parent context, if any
    :var Signal resource_added: a signal (:class:`ResourceEvent`) dispatched when a resource
        has been published in this context
    """

    resource_added = Signal(ResourceEvent)

    _loop: AbstractEventLoop | None = None
    _reset_token: Token

    def __init__(self, parent: Optional[Context] = None) -> None:
        assert check_argument_types()
        if parent is None:
            self._parent = _current_context.get(None)
        else:
            warnings.warn(
                "Explicitly passing the parent context has been deprecated. "
                "The context stack is now tracked by the means of PEP 555 context "
                "variables.",
                DeprecationWarning,
                stacklevel=2,
            )
            self._parent = parent

        self._state = ContextState.open
        self._resources: Dict[Tuple[type, str], ResourceContainer] = {}
        self._resource_factories: Dict[Tuple[type, str], ResourceContainer] = {}
        self._resource_factories_by_context_attr: Dict[str, ResourceContainer] = {}
        self._teardown_callbacks: List[Tuple[Callable, bool]] = []

    def __getattr__(self, name):
        # First look for a resource factory in the whole context chain
        for ctx in self.context_chain:
            factory = ctx._resource_factories_by_context_attr.get(name)
            if factory:
                return factory.generate_value(self)

        # When that fails, look directly for an attribute in the parents
        for ctx in self.context_chain[1:]:
            value = getattr_static(ctx, name, None)
            if value is not None:
                return getattr(ctx, name)

        raise AttributeError(f"no such context variable: {name}")

    @property
    def context_chain(self) -> List[Context]:
        """Return a list of contexts starting from this one, its parent and so on."""
        contexts = []
        ctx: Optional[Context] = self
        while ctx is not None:
            contexts.append(ctx)
            ctx = ctx.parent

        return contexts

    @property
    def loop(self) -> AbstractEventLoop:
        """Return the event loop associated with this context."""
        if self._loop is None:
            self._loop = get_running_loop()

        return self._loop

    @property
    def parent(self) -> Optional[Context]:
        """Return the parent context, or ``None`` if there is no parent."""
        return self._parent

    @property
    def closed(self) -> bool:
        """
        Return ``True`` if the teardown process has at least been initiated, ``False``
        otherwise.

        """
        return self._state is not ContextState.open

    def _check_closed(self) -> None:
        if self._state is ContextState.closed:
            raise RuntimeError("this context has already been closed")

    def add_teardown_callback(
        self, callback: Callable, pass_exception: bool = False
    ) -> None:
        """
        Add a callback to be called when this context closes.

        This is intended for cleanup of resources, and the list of callbacks is processed in the
        reverse order in which they were added, so the last added callback will be called first.

        The callback may return an awaitable. If it does, the awaitable is awaited on before
        calling any further callbacks.

        :param callback: a callable that is called with either no arguments or with the exception
            that ended this context, based on the value of ``pass_exception``
        :param pass_exception: ``True`` to pass the callback the exception that ended this context
            (or ``None`` if the context ended cleanly)

        """
        assert check_argument_types()
        self._check_closed()
        self._teardown_callbacks.append((callback, pass_exception))

    async def close(self, exception: BaseException | None = None) -> None:
        """
        Close this context and call any necessary resource teardown callbacks.

        If a teardown callback returns an awaitable, the return value is awaited on before calling
        any further teardown callbacks.

        All callbacks will be processed, even if some of them raise exceptions. If at least one
        callback raised an error, this method will raise a :exc:`~.TeardownError` at the end.

        After this method has been called, resources can no longer be requested or published on
        this context.

        :param exception: the exception, if any, that caused this context to be closed
        :raises .TeardownError: if one or more teardown callbacks raise an exception

        """
        self._check_closed()
        if self._state is ContextState.closing:
            raise RuntimeError("this context is already closing")

        self._state = ContextState.closing

        try:
            exceptions = []
            while self._teardown_callbacks:
                callbacks, self._teardown_callbacks = self._teardown_callbacks, []
                for callback, pass_exception in reversed(callbacks):
                    try:
                        retval = callback(exception) if pass_exception else callback()
                        if isawaitable(retval):
                            await retval
                    except Exception as e:
                        exceptions.append(e)

            if exceptions:
                raise TeardownError(exceptions)
        finally:
            self._state = ContextState.closed

    def __enter__(self):
        warnings.warn(
            "Using Context as a synchronous context manager has been deprecated",
            DeprecationWarning,
        )
        self._check_closed()

        if self._loop is None:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", DeprecationWarning)
                self._loop = get_event_loop()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.loop.run_until_complete(self.close(exc_val))

    async def __aenter__(self) -> Context:
        self._check_closed()
        if self._loop is None:
            self._loop = get_running_loop()

        self._host_task = current_task()
        self._reset_token = _current_context.set(self)
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> None:
        try:
            await self.close(exc_val)
        finally:
            try:
                _current_context.reset(self._reset_token)
            except ValueError:
                warnings.warn(
                    f"Potential context stack corruption detected. This context "
                    f"({hex(id(self))}) was entered in task {self._host_task} and exited "
                    f"in task {current_task()}. If this happened because you entered "
                    f"the context in an async generator, you should try to defer that to a "
                    f"regular async function."
                )

    def add_resource(
        self,
        value,
        name: str = "default",
        context_attr: str | None = None,
        types: Union[type, Sequence[type]] = (),
    ) -> None:
        """
        Add a resource to this context.

        This will cause a ``resource_added`` event to be dispatched.

        :param value: the actual resource value
        :param name: name of this resource (unique among all its registered types within
            a single context)
        :param context_attr: (deprecated) name of the context attribute this resource
            will be accessible as
        :param types: type(s) to register the resource as (omit to use the type of
            ``value``)
        :raises asphalt.core.context.ResourceConflict: if the resource conflicts with an
            existing one in any way

        """
        # TODO: re-enable when typeguard properly identifies parametrized types as types
        # assert check_argument_types()
        self._check_closed()
        if types:
            if (
                isclass(types)
                or get_origin(types) is not None
                or not isinstance(types, ABCSequence)
            ):
                types = (cast(type, types),)

            if not all(isclass(x) or get_origin(x) is not None for x in types):
                raise TypeError("types must be a type or sequence of types")
        else:
            types = (type(value),)

        if value is None:
            raise ValueError('"value" must not be None')
        if not resource_name_re.fullmatch(name):
            raise ValueError(
                '"name" must be a nonempty string consisting only of alphanumeric '
                "characters and underscores"
            )
        if context_attr and getattr_static(self, context_attr, None) is not None:
            raise ResourceConflict(
                f"this context already has an attribute {context_attr!r}"
            )

        for resource_type in types:
            if (resource_type, name) in self._resources:
                raise ResourceConflict(
                    f"this context already contains a resource of type "
                    f"{qualified_name(resource_type)} using the name {name!r}"
                )

        resource = ResourceContainer(value, tuple(types), name, context_attr, False)
        for type_ in resource.types:
            self._resources[(type_, name)] = resource

        if context_attr:
            warnings.warn(
                "context attributes have been deprecated in favor of dependency "
                "injection",
                DeprecationWarning,
                stacklevel=2,
            )
            setattr(self, context_attr, value)

        # Notify listeners that a new resource has been made available
        self.resource_added.dispatch(types, name, False)

    def add_resource_factory(
        self,
        factory_callback: factory_callback_type,
        types: Union[type, Sequence[Type], None] = None,
        name: str = "default",
        context_attr: str | None = None,
    ) -> None:
        """
        Add a resource factory to this context.

        This will cause a ``resource_added`` event to be dispatched.

        A resource factory is a callable that generates a "contextual" resource when it is
        requested by either using any of the methods :meth:`get_resource`, :meth:`require_resource`
        or :meth:`request_resource` or its context attribute is accessed.

        The type(s) of the generated resources need to be specified, either by passing
        the ``types`` argument, or by adding a return type annotation to the factory
        function. If the generated resource needs to be registered as multiple types,
        you can use :data:`~typing.Union` (e.g. ``Union[str, int]``) or a union type
        (e.g. ``str | int``; requires ``from __future__ import annotations`` on earlier
        than Python 3.10).

        When a new resource is created in this manner, it is always bound to the context through
        it was requested, regardless of where in the chain the factory itself was added to.

        :param factory_callback: a (non-coroutine) callable that takes a context instance as
            argument and returns the created resource object
        :param types: one or more types to register the generated resource as on the target context
            (can be omitted if the factory callable has a return type annotation)
        :param name: name of the resource that will be created in the target context
        :param context_attr: name of the context attribute the created resource will be accessible
            as
        :raises asphalt.core.context.ResourceConflict: if there is an existing resource factory for
            the given type/name combinations or the given context variable

        """
        import types as stdlib_types

        self._check_closed()
        if not resource_name_re.fullmatch(name):
            raise ValueError(
                '"name" must be a nonempty string consisting only of alphanumeric '
                "characters and underscores"
            )
        if iscoroutinefunction(factory_callback):
            raise TypeError('"factory_callback" must not be a coroutine function')

        if types is not None:
            if isinstance(types, type):
                resource_types: Tuple[type, ...] = (types,)
            else:
                resource_types = tuple(types)
        else:
            # Extract the resources types from the return type annotation of the factory
            type_hints = get_type_hints(factory_callback)
            try:
                return_type_hint = type_hints["return"]
            except KeyError:
                raise ValueError(
                    "no resource types specified, and the factory callback does not "
                    "have a return type hint"
                )

            origin = get_origin(return_type_hint)
            if origin is Union or (
                sys.version_info >= (3, 10) and origin is stdlib_types.UnionType
            ):
                resource_types = return_type_hint.__args__
            else:
                resource_types = (return_type_hint,)

        if not resource_types:
            raise ValueError("no resource types were specified")

        if None in resource_types:
            raise TypeError("None is not a valid resource type")

        # Check for a conflicting context attribute
        if context_attr in self._resource_factories_by_context_attr:
            raise ResourceConflict(
                f"this context already contains a resource factory for the context attribute "
                f"{context_attr!r}"
            )

        # Check for conflicts with existing resource factories
        for type_ in resource_types:
            if (type_, name) in self._resource_factories:
                raise ResourceConflict(
                    "this context already contains a resource factory for the "
                    "type {}".format(qualified_name(type_))
                )

        # Add the resource factory to the appropriate lookup tables
        resource = ResourceContainer(
            factory_callback, resource_types, name, context_attr, True
        )
        for type_ in resource_types:
            self._resource_factories[(type_, name)] = resource

        if context_attr:
            warnings.warn(
                "context attributes have been deprecated in favor of dependency "
                "injection",
                DeprecationWarning,
                stacklevel=2,
            )
            self._resource_factories_by_context_attr[context_attr] = resource

        # Notify listeners that a new resource has been made available
        self.resource_added.dispatch(resource_types, name, True)

    def get_resource(
        self, type: Type[T_Resource], name: str = "default"
    ) -> Optional[T_Resource]:
        """
        Look up a resource in the chain of contexts.

        :param type: type of the requested resource
        :param name: name of the requested resource
        :return: the requested resource, or ``None`` if none was available

        """
        # TODO: re-enable when typeguard properly identifies parametrized types as types
        # assert check_argument_types()
        self._check_closed()
        key = (type, name)

        # First check if there's already a matching resource in this context
        resource = self._resources.get(key)
        if resource is not None:
            return resource.value_or_factory

        # Next, check if there's a resource factory available on the context chain
        resource = next(
            (
                ctx._resource_factories[key]
                for ctx in self.context_chain
                if key in ctx._resource_factories
            ),
            None,
        )
        if resource is not None:
            return resource.generate_value(self)

        # Finally, check parents for a matching resource
        return next(
            (
                ctx._resources[key].value_or_factory
                for ctx in self.context_chain
                if key in ctx._resources
            ),
            None,
        )

    def get_resources(self, type: Type[T_Resource]) -> Set[T_Resource]:
        """
        Retrieve all the resources of the given type in this context and its parents.

        Any matching resource factories are also triggered if necessary.

        :param type: type of the resources to get
        :return: a set of all found resources of the given type

        """
        assert check_argument_types()

        # Collect all the matching resources from this context
        resources: Dict[str, T_Resource] = {
            container.name: container.value_or_factory
            for container in self._resources.values()
            if not container.is_factory and type in container.types
        }

        # Next, find all matching resource factories in the context chain and generate resources
        resources.update(
            {
                container.name: container.generate_value(self)
                for ctx in self.context_chain
                for container in ctx._resources.values()
                if container.is_factory
                and type in container.types
                and container.name not in resources
            }
        )

        # Finally, add the resource values from the parent contexts
        resources.update(
            {
                container.name: container.value_or_factory
                for ctx in self.context_chain[1:]
                for container in ctx._resources.values()
                if not container.is_factory
                and type in container.types
                and container.name not in resources
            }
        )

        return set(resources.values())

    def require_resource(
        self, type: Type[T_Resource], name: str = "default"
    ) -> T_Resource:
        """
        Look up a resource in the chain of contexts and raise an exception if it is not found.

        This is like :meth:`get_resource` except that instead of returning ``None`` when a resource
        is not found, it will raise :exc:`~asphalt.core.context.ResourceNotFound`.

        :param type: type of the requested resource
        :param name: name of the requested resource
        :return: the requested resource
        :raises asphalt.core.context.ResourceNotFound: if a resource of the given type and name was
            not found

        """
        resource = self.get_resource(type, name)
        if resource is None:
            raise ResourceNotFound(type, name)

        return resource

    async def request_resource(
        self, type: Type[T_Resource], name: str = "default"
    ) -> T_Resource:
        """
        Look up a resource in the chain of contexts.

        This is like :meth:`get_resource` except that if the resource is not already available, it
        will wait for one to become available.

        :param type: type of the requested resource
        :param name: name of the requested resource
        :return: the requested resource

        """
        # First try to locate an existing resource in this context and its parents
        value = self.get_resource(type, name)
        if value is not None:
            return value

        # Wait until a matching resource or resource factory is available
        signals = [ctx.resource_added for ctx in self.context_chain]
        await wait_event(
            signals,
            lambda event: event.resource_name == name and type in event.resource_types,
        )
        return self.require_resource(type, name)

    def call_async(self, func: Callable[..., T_Retval], *args, **kwargs) -> T_Retval:
        """
        Call the given callable in the event loop thread.

        This method lets you call asynchronous code from a worker thread.
        Do not use it from within the event loop thread.

        If the callable returns an awaitable, it is resolved before returning to the caller.

        :param func: a regular function or a coroutine function
        :param args: positional arguments to call the callable with
        :param kwargs: keyword arguments to call the callable with
        :return: the return value of the call

        """
        return asyncio_extras.call_async(self.loop, func, *args, **kwargs)

    def call_in_executor(
        self,
        func: Callable[..., T_Retval],
        *args,
        executor: Union[Executor, str] = None,
        **kwargs,
    ) -> Awaitable[T_Retval]:
        """
        Call the given callable in an executor.

        :param func: the callable to call
        :param args: positional arguments to call the callable with
        :param executor: either an :class:`~concurrent.futures.Executor` instance, the resource
            name of one or ``None`` to use the event loop's default executor
        :param kwargs: keyword arguments to call the callable with
        :return: an awaitable that resolves to the return value of the call

        """
        assert check_argument_types()
        if isinstance(executor, str):
            executor = self.require_resource(Executor, executor)

        # Fill in self._loop if it's None
        if self._loop is None:
            self._loop = get_running_loop()

        callback: partial[T_Retval] = partial(copy_context().run, func, *args, **kwargs)
        return self._loop.run_in_executor(executor, callback)

    def threadpool(self, executor: Union[Executor, str, None] = None):
        """
        Return an asynchronous context manager that runs the block in a (thread pool) executor.

        :param executor: either an :class:`~concurrent.futures.Executor` instance, the resource
            name of one or ``None`` to use the event loop's default executor
        :return: an asynchronous context manager

        """
        assert check_argument_types()
        if isinstance(executor, str):
            executor = self.require_resource(Executor, executor)

        return asyncio_extras.threadpool(executor)


def executor(arg: Union[Executor, str, Callable] = None):
    """
    Decorate a function so that it runs in an :class:`~concurrent.futures.Executor`.

    If a resource name is given, the first argument must be a :class:`~.Context`.

    Usage::

        @executor
        def should_run_in_executor():
            ...

    With a resource name::

        @executor('resourcename')
        def should_run_in_executor(ctx):
            ...

    :param arg: a callable to decorate, an :class:`~concurrent.futures.Executor` instance, the
        resource name of one or ``None`` to use the event loop's default executor
    :return: the wrapped function

    """

    def outer_wrapper(func: Callable):
        @wraps(func)
        def inner_wrapper(*args, **kwargs):
            try:
                ctx = next(arg for arg in args[:2] if isinstance(arg, Context))
            except StopIteration:
                raise RuntimeError(
                    "the first positional argument to {}() has to be a Context "
                    "instance".format(callable_name(func))
                ) from None

            executor = ctx.require_resource(Executor, resource_name)
            return asyncio_extras.call_in_executor(
                func, *args, executor=executor, **kwargs
            )

        return inner_wrapper

    if isinstance(arg, str):
        resource_name = arg
        return outer_wrapper

    return asyncio_extras.threadpool(arg)


@overload
def context_teardown(
    func: Callable[[T_Context], AsyncGenerator[None, Exception | None]]
) -> Callable[[T_Context], Coroutine[Any, Any, None]]:
    ...


@overload
def context_teardown(
    func: Callable[[T_Self, T_Context], AsyncGenerator[None, Exception | None]]
) -> Callable[[T_Self, T_Context], Coroutine[Any, Any, None]]:
    ...


def context_teardown(
    func: Callable[[T_Context], AsyncGenerator[None, Exception | None]]
    | Callable[[T_Self, T_Context], AsyncGenerator[None, Exception | None]]
) -> Callable[[T_Context], Coroutine[Any, Any, None]] | Callable[
    [T_Self, T_Context], Coroutine[Any, Any, None]
]:
    """
    Wrap an async generator function to execute the rest of the function at context teardown.

    This function returns an async function, which, when called, starts the wrapped async
    generator. The wrapped async function is run until the first ``yield`` statement
    When the context is being torn down, the exception that ended the context, if any, is sent to
    the generator.

    For example::

        class SomeComponent(Component):
            @context_teardown
            async def start(self, ctx: Context):
                service = SomeService()
                ctx.add_resource(service)
                exception = yield
                service.stop()

    :param func: an async generator function
    :return: an async function

    """

    @wraps(func)
    async def wrapper(*args, **kwargs) -> None:
        async def teardown_callback(exception: Optional[Exception]) -> None:
            try:
                await generator.asend(exception)
            except StopAsyncIteration:
                pass
            finally:
                await generator.aclose()

        try:
            ctx = next(arg for arg in args[:2] if isinstance(arg, Context))
        except StopIteration:
            raise RuntimeError(
                "the first positional argument to {}() has to be a Context "
                "instance".format(callable_name(func))
            ) from None

        generator = func(*args, **kwargs)
        try:
            await generator.asend(None)
        except StopAsyncIteration:
            pass
        except BaseException:
            await generator.aclose()
            raise
        else:
            ctx.add_teardown_callback(teardown_callback, True)

    if not isasyncgenfunction(func):
        if async_generator and iscoroutinefunction(func):
            warnings.warn(
                "Using @context_teardown on regular coroutine functions has been "
                "deprecated",
                DeprecationWarning,
                stacklevel=2,
            )
            func = async_generator(func)
        else:
            raise TypeError(
                f"{callable_name(func)} must be an async generator function"
            )

    return wrapper


def current_context() -> Context:
    """
    Return the currently active context.

    :raises NoCurrentContext: if there is no active context

    """
    ctx = _current_context.get()
    if ctx is None:
        raise NoCurrentContext

    return ctx


def get_resource(type: Type[T_Resource], name: str = "default") -> T_Resource | None:
    """Shortcut for ``current_context().get_resource(...)``."""
    return current_context().get_resource(type, name)


def get_resources(type: Type[T_Resource]) -> set[T_Resource]:
    """Shortcut for ``current_context().get_resources(...)``."""
    return current_context().get_resources(type)


def require_resource(type: Type[T_Resource], name: str = "default") -> T_Resource:
    """Shortcut for ``current_context().require_resource(...)``."""
    return current_context().require_resource(type, name)


@dataclass
class _Dependency:
    name: str = "default"
    cls: type = field(init=False)
    optional: bool = field(init=False, default=False)

    def __getattr__(self, item):
        raise AttributeError(
            "Attempted to access an attribute in a resource() marker – did you forget "
            "to add the @inject decorator?"
        )


def resource(name: str = "default") -> Any:
    """
    Marker for declaring a parameter for dependency injection via :func:`inject`.

    :param name: the resource name (defaults to ``default``)

    """
    return _Dependency(name)


def Dependency(name: str = "default") -> Any:
    warnings.warn(
        "Dependency() has been deprecated (use resource() instead)",
        DeprecationWarning,
        stacklevel=2,
    )
    return resource(name)


@overload
def inject(
    func: Callable[P, Coroutine[Any, Any, T_Retval]]
) -> Callable[P, Coroutine[Any, Any, T_Retval]]:
    ...


@overload
def inject(func: Callable[P, T_Retval]) -> Callable[P, T_Retval]:
    ...


def inject(func: Callable[P, Any]) -> Callable[P, Any]:
    """
    Wrap the given coroutine function for use with dependency injection.

    Parameters with dependencies need to be annotated and have :func:`resource` as the
    default value. When the wrapped function is called, values for such parameters will
    be automatically filled in by calling :func:`require_resource` using the parameter's
    type annotation and the resource name passed to :func:`resource` (or ``"default"``)
    as the arguments.

    Any forward references among the type annotations are resolved on the first call to
    the wrapped function.

    """
    forward_refs_resolved = False
    local_names = sys._getframe(1).f_locals if "<locals>" in func.__qualname__ else {}

    def resolve_forward_refs() -> None:
        nonlocal forward_refs_resolved, local_names
        type_hints = get_type_hints(func, localns=local_names)
        for key, dependency in injected_resources.items():
            dependency.cls = type_hints[key]
            origin = get_origin(type_hints[key])
            if origin is Union or (
                sys.version_info >= (3, 10) and origin is types.UnionType  # noqa: E721
            ):
                args = [
                    arg
                    for arg in get_args(dependency.cls)
                    if arg is not type(None)  # noqa: E721
                ]
                if len(args) == 1:
                    dependency.optional = True
                    dependency.cls = args[0]
                else:
                    raise TypeError(
                        "Unions are only valid with dependency injection when there "
                        "are exactly two items and other item is None"
                    )

        del local_names
        forward_refs_resolved = True

    @wraps(func)
    def sync_wrapper(*args, **kwargs) -> T_Retval:
        if not forward_refs_resolved:
            resolve_forward_refs()

        ctx = current_context()
        resources: dict[str, Any] = {}
        for argname, dependency in injected_resources.items():
            if dependency.optional:
                resource = ctx.get_resource(dependency.cls, dependency.name)
            else:
                resource = ctx.require_resource(dependency.cls, dependency.name)

            resources[argname] = resource

        return func(*args, **kwargs, **resources)

    @wraps(func)
    async def async_wrapper(*args, **kwargs) -> T_Retval:
        if not forward_refs_resolved:
            resolve_forward_refs()

        ctx = current_context()
        resources: dict[str, Any] = {}
        for argname, dependency in injected_resources.items():
            if dependency.optional:
                resource = ctx.get_resource(dependency.cls, dependency.name)
            else:
                resource = ctx.require_resource(dependency.cls, dependency.name)

            resources[argname] = resource

        return await func(*args, **kwargs, **resources)

    sig = signature(func)
    injected_resources: dict[str, _Dependency] = {}
    for param in sig.parameters.values():
        if isinstance(param.default, _Dependency):
            if param.kind is Parameter.POSITIONAL_ONLY:
                raise TypeError(
                    f"Cannot inject dependency to positional-only parameter "
                    f"{param.name!r}"
                )

            if param.annotation is Parameter.empty:
                raise TypeError(
                    f"Dependency for parameter {param.name!r} of function "
                    f"{callable_name(func)!r} is missing the type annotation"
                )

            injected_resources[param.name] = param.default
        elif param.default is resource:
            raise TypeError(
                f"Default value for parameter {param.name!r} of function "
                f"{callable_name(func)} was the 'resource' function – did you forget "
                f"to add the parentheses at the end?"
            )

    if injected_resources:
        if iscoroutinefunction(func):
            return async_wrapper
        else:
            return sync_wrapper
    else:
        warnings.warn(
            f"{callable_name(func)} does not have any injectable resources declared"
        )
        return func
